from collections.abc import Callable
from copy import deepcopy
from os import cpu_count
from time import time

import numpy as np
from joblib import Parallel, delayed

from .io import json_write
from .utilities import last_n_elements_equal


# PSO class definition with genetic operators
class COBRAKPAPSO:
    def __init__(
        self,
        fitness_function: Callable[
            [list[float | int]], tuple[float, list[float | int]]
        ],
        xs_dim: int,
        gen: int,
        omega: float = 0.7298,
        eta1: float = 2.05,
        eta2: float = 2.05,
        max_vel: float = 0.5,
        extra_xs: list[list[float]] = [],
        seed: int | None = None,
        objvalue_json_path: str = "",
        max_rounds_same_objvalue: float = float("inf"),
        pop_size: int | None = None,
        diversity_threshold: float = 0.1,  # New parameter for diversity threshold
        diversity_restore_factor: float = 0.1,  # New parameter for diversity restoration
        ga_crossover_rate: float = 0.7,  # Crossover rate for genetic algorithm
        ga_mutation_rate: float = 0.01,  # Mutation rate for genetic algorithm
    ) -> None:
        # Parameters
        self.fitness_function = fitness_function
        self.xs_dim = xs_dim
        self.gen = gen
        self.omega = omega
        self.eta1 = eta1
        self.eta2 = eta2
        self.max_vel = max_vel
        self.seed = seed
        if seed is not None:
            np.random.seed(seed)

        # Initialization of random particles
        used_cpu_count = cpu_count() if pop_size is None else pop_size
        if used_cpu_count is None:
            used_cpu_count = 1
        self.xs = [
            np.random.uniform(0.0, 1.0, size=xs_dim).tolist()
            for _ in range(used_cpu_count - len(extra_xs))
        ]

        # Addition of user-defined extra particles
        if extra_xs != []:
            self.xs.extend(deepcopy(extra_xs))

        self.velocities = [
            np.random.uniform(-self.max_vel, self.max_vel, size=xs_dim).tolist()
            for _ in range(len(self.xs))
        ]
        self.personal_best_xs = self.xs.copy()
        self.personal_best_fs = [float("inf")] * len(self.xs)
        self.objvalue_json_path = objvalue_json_path
        self.objvalue_json_data: dict[float, list[float]] = {}
        self.max_rounds_same_objvalue = max_rounds_same_objvalue
        self.diversity_threshold = diversity_threshold
        self.diversity_restore_factor = diversity_restore_factor
        self.ga_crossover_rate = ga_crossover_rate
        self.ga_mutation_rate = ga_mutation_rate

    def run(self) -> tuple[float, list[float]]:
        with Parallel(n_jobs=-1) as parallel:
            # Initialization
            particle_xs = self.xs.copy()
            print("INIT START")
            particle_fs = parallel(
                delayed(self.fitness_function)(x) for x in particle_xs
            )
            if particle_fs is None:
                print("ERROR: Something went wrong during initialization!")
                raise ValueError
            particle_fs = [particle_fs[i][0] for i in range(len(particle_fs))]
            print("INIT END")
            best_swarm_x = particle_xs[particle_fs.index(min(particle_fs))].copy()
            best_f = min(particle_fs)

            # Update personal bests
            for i in range(len(particle_fs)):
                self.personal_best_fs[i] = particle_fs[i]
                self.personal_best_xs[i] = particle_xs[i].copy()

            if self.objvalue_json_path != "":
                start_time = time()
                self.objvalue_json_data[0.0] = deepcopy(self.personal_best_fs)
                json_write(self.objvalue_json_path, self.objvalue_json_data)

            # Actual algorithm
            max_objvalues = []
            for gen in range(self.gen):
                max_objvalues.append(max(self.personal_best_fs))
                if last_n_elements_equal(max_objvalues, self.max_rounds_same_objvalue):
                    break

                # Check diversity and restore if necessary
                if self.check_diversity(particle_xs) < self.diversity_threshold:
                    self.restore_diversity(particle_xs)

                # Genetic operators
                self.apply_genetic_operators(particle_xs, particle_fs)

                # Update velocities and positions in parallel
                results = parallel(
                    delayed(self.update_particle)(
                        self.personal_best_xs[current_particle],
                        particle_xs[current_particle],
                        self.velocities[current_particle],
                        best_swarm_x,
                    )
                    for current_particle in range(len(self.xs))
                )
                if results is None:
                    print("ERROR: Something went wrong during fitness calculations!")
                    raise ValueError

                # Unpack results
                for current_particle, (x, velocity, fitness) in enumerate(results):
                    particle_xs[current_particle] = x
                    self.velocities[current_particle] = velocity
                    particle_fs[current_particle] = fitness

                    # Update personal best
                    if (
                        particle_fs[current_particle]
                        < self.personal_best_fs[current_particle]
                    ):
                        self.personal_best_fs[current_particle] = particle_fs[
                            current_particle
                        ]
                        self.personal_best_xs[current_particle] = x.copy()

                    # Update the best swarm position if necessary
                    if particle_fs[current_particle] < best_f:
                        best_swarm_x = x.copy()
                        best_f = particle_fs[current_particle]

                if self.objvalue_json_path != "":
                    self.objvalue_json_data[time() - start_time] = deepcopy(
                        self.personal_best_fs
                    )
                    json_write(self.objvalue_json_path, self.objvalue_json_data)

        return best_f, best_swarm_x

    def update_particle(
        self,
        personal_best_x: list[float],
        old_x: list[float],
        velocity: list[float],
        best_swarm_x: list[float],
    ) -> tuple[list[float], list[float], float]:
        phi1 = np.random.uniform(0.0, 1.0, size=self.xs_dim)
        phi2 = np.random.uniform(0.0, 1.0, size=self.xs_dim)

        # Update velocity
        velocity = (
            self.omega * np.array(velocity)
            + self.eta1 * phi1 * (np.array(personal_best_x) - np.array(old_x))
            + self.eta2 * phi2 * (np.array(best_swarm_x) - np.array(old_x))
        ).tolist()

        # Clamp velocity
        velocity = [max(min(v, self.max_vel), -self.max_vel) for v in velocity]

        # Update position
        x = (np.array(old_x) + np.array(velocity)).tolist()

        # Boundary handling (reflection)
        x = [max(min(x_i, 1.0), 0.0) if x_i < 0.0 or x_i > 1.0 else x_i for x_i in x]

        # Update particle fitness
        fitness, _ = self.fitness_function(x)

        return x, velocity, fitness

    def check_diversity(self, particle_xs: list[list[float]]) -> float:
        # Calculate the average distance between particles
        distances = []
        for i in range(len(particle_xs)):
            for j in range(i + 1, len(particle_xs)):
                dist = np.linalg.norm(
                    np.array(particle_xs[i]) - np.array(particle_xs[j])
                )
                distances.append(dist)
        return np.mean(distances) if distances else 0.0

    def restore_diversity(self, particle_xs: list[list[float]]) -> None:
        # Perturb particles to restore diversity
        for i in range(len(particle_xs)):
            perturbation = np.random.uniform(
                -self.diversity_restore_factor,
                self.diversity_restore_factor,
                size=self.xs_dim,
            )
            particle_xs[i] = (np.array(particle_xs[i]) + perturbation).tolist()
            # Boundary handling (reflection)
            particle_xs[i] = [
                max(min(x_i, 1.0), 0.0) if x_i < 0.0 or x_i > 1.0 else x_i
                for x_i in particle_xs[i]
            ]

    def apply_genetic_operators(
        self, particle_xs: list[list[float]], particle_fs: list[float]
    ) -> None:
        # Selection: Select particles based on fitness
        sorted_indices = np.argsort(particle_fs)
        selected_indices = sorted_indices[: len(particle_xs) // 2]

        # Crossover: Combine selected particles to create new offspring
        offspring = []
        for i in range(0, len(selected_indices) - 1, 2):
            parent1 = particle_xs[selected_indices[i]]
            parent2 = particle_xs[selected_indices[i + 1]]
            if np.random.rand() < self.ga_crossover_rate:
                child1, child2 = self.crossover(parent1, parent2)
                offspring.extend((child1, child2))
            else:
                offspring.extend((parent1, parent2))

        # Mutation: Introduce random changes to some particles
        for i in range(len(offspring)):
            if np.random.rand() < self.ga_mutation_rate:
                offspring[i] = self.mutation(offspring[i])

        # Replace some particles with offspring
        for i in range(len(offspring)):
            particle_xs[i] = offspring[i]

    def crossover(
        self, parent1: list[float], parent2: list[float]
    ) -> tuple[list[float], list[float]]:
        # Single-point crossover
        crossover_point = np.random.randint(1, self.xs_dim - 1)
        child1 = parent1[:crossover_point] + parent2[crossover_point:]
        child2 = parent2[:crossover_point] + parent1[crossover_point:]
        return child1, child2

    def mutation(self, particle: list[float]) -> list[float]:
        # Random mutation
        mutated_particle = particle.copy()
        for i in range(self.xs_dim):
            if np.random.rand() < self.ga_mutation_rate:
                mutated_particle[i] += np.random.uniform(
                    -self.ga_mutation_rate, self.ga_mutation_rate
                )
                mutated_particle[i] = max(
                    min(mutated_particle[i], 1.0), 0.0
                )  # Boundary handling
        return mutated_particle



from copy import deepcopy
from random import sample
from tempfile import TemporaryDirectory
from time import time
from typing import Any, Literal

from joblib import Parallel, cpu_count, delayed
from numpy.random import randint
from pyomo.common.errors import ApplicationError

from .constants import ALL_OK_KEY, OBJECTIVE_VAR_NAME, Z_VAR_PREFIX
from .dataclasses import CorrectionConfig, ExtraLinearConstraint, Model, Solver
from .genetic import COBRAKGENETIC
from .io import json_load, json_write
from .lps import perform_lp_optimization, perform_lp_variability_analysis
from .nlps import perform_nlp_irreversible_optimization_with_active_reacs_only
from .pso_parallel import COBRAKPAPSO
from .standard_solvers import IPOPT, SCIP
from .utilities import (
    get_active_reacs_from_optimization_dict,
    get_files,
    get_stoichiometrically_coupled_reactions,
    is_objsense_maximization,
    split_list,
    standardize_folder,
)


class COBRAKProblem:
    def __init__(
        self,
        cobrak_model: Model,
        objective_target: dict[str, float],
        objective_sense: int,
        variability_dict: dict[str, tuple[float, float]],
        nlp_dict_list: list[dict[str, float]],
        best_value: float,
        with_kappa: bool = True,
        with_gamma: bool = True,
        with_iota: bool = True,
        with_alpha: bool = True,
        num_gens: int = 5,
        algorithm: Literal["pso", "genetic"] = "pso",
        lp_solver: Solver = SCIP,
        nlp_solver: Solver = IPOPT,
        objvalue_json_path: str = "",
        max_rounds_same_objvalue: float = float("inf"),
        correction_config: CorrectionConfig = CorrectionConfig(),
        min_abs_objvalue: float = 1e-6,
        pop_size: int | None = None,
    ) -> None:
        self.original_cobrak_model: Model = deepcopy(cobrak_model)
        self.objective_target = objective_target
        self.objective_sense = objective_sense
        self.blocked_reacs: list[str] = []
        self.initial_xs_list: list[list[int | float]] = [
            [] for _ in range(len(nlp_dict_list))
        ]
        self.minimal_xs_dict: dict[float, list[float]] = {}
        self.variability_data = deepcopy(variability_dict)
        self.algorithm = algorithm

        reac_couples = get_stoichiometrically_coupled_reactions(
            self.original_cobrak_model
        )

        objective_target_ids = list(objective_target.keys())
        filtered_reac_couples: list[tuple[str, ...]] = []
        for reac_couple in reac_couples:
            filtered_reac_couple = [
                reac_id
                for reac_id in reac_couple
                if (abs(self.variability_data[reac_id][1]) > 0.0)
                and (abs(self.variability_data[reac_id][0]) <= 0.0)
                and not (
                    cobrak_model.reactions[reac_id].dG0 is None
                    and cobrak_model.reactions[reac_id].enzyme_reaction_data is None
                )
            ]

            found_invalid_id = False
            for objective_target_id in objective_target_ids:
                if objective_target_id in filtered_reac_couple:
                    found_invalid_id = True
            for var_id in correction_config.error_scenario:
                if var_id in filtered_reac_couple:
                    found_invalid_id = True

            if found_invalid_id:
                continue

            if len(filtered_reac_couple) > 0:
                filtered_reac_couples.append(tuple(filtered_reac_couple))

        self.idx_to_reac_ids: dict[int, tuple[str, ...]] = {}
        couple_idx = 0
        for filtered_reac_couplex in filtered_reac_couples:
            nlp_idx = 0
            for nlp_dict in nlp_dict_list:
                first_reac_id = filtered_reac_couplex[0]
                if (first_reac_id in nlp_dict) and (nlp_dict[first_reac_id] > 0.0):
                    self.initial_xs_list[nlp_idx].append(
                        1.0 if algorithm == "pso" else 1
                    )
                else:
                    self.initial_xs_list[nlp_idx].append(
                        0.0 if algorithm == "pso" else 0
                    )

                nlp_idx += 1  # noqa: SIM113

            self.idx_to_reac_ids[couple_idx] = filtered_reac_couplex
            couple_idx += 1

        self.with_kappa = with_kappa
        self.with_gamma = with_gamma
        self.with_iota = with_iota
        self.with_alpha = with_alpha
        self.dim = couple_idx
        self.num_gens = num_gens
        self.lp_solver = lp_solver
        self.nlp_solver = nlp_solver
        self.temp_directory_name = ""
        self.best_value = best_value
        self.objvalue_json_path = objvalue_json_path
        self.max_rounds_same_objvalue = max_rounds_same_objvalue
        self.correction_config = correction_config
        self.min_abs_objvalue = min_abs_objvalue
        self.pop_size = pop_size

    def fitness(
        self,
        x: list[float | int],
    ) -> list[tuple[float, list[float | int]]]:
        # Preliminary TFBA :3
        deactivated_reactions: list[str] = []
        for couple_idx, reac_ids in self.idx_to_reac_ids.items():
            if x[couple_idx] <= 0.02:
                deactivated_reactions.extend(reac_ids)

        try:
            first_ectfba_dict = perform_lp_optimization(
                cobrak_model=self.original_cobrak_model,
                objective_target=self.objective_target,
                objective_sense=self.objective_sense,
                with_enzyme_constraints=True,
                with_thermodynamic_constraints=True,
                with_loop_constraints=True,
                variability_dict=deepcopy(self.variability_data),
                ignored_reacs=deactivated_reactions,
                solver=self.lp_solver,
                correction_config=self.correction_config,
            )
        except (ApplicationError, AttributeError, ValueError):
            first_ectfba_dict = {ALL_OK_KEY: False}
        if not first_ectfba_dict[ALL_OK_KEY]:
            return [(1_000_000.0, [])]

        nlp_results: list[dict[str, float]] = []

        if is_objsense_maximization(self.objective_sense):
            lower_value = first_ectfba_dict[OBJECTIVE_VAR_NAME] - 1e-12
            upper_value = None
        else:
            lower_value = None
            upper_value = first_ectfba_dict[OBJECTIVE_VAR_NAME] + 1e-12

        maxz_model = deepcopy(self.original_cobrak_model)
        maxz_model.extra_linear_constraints = [
            ExtraLinearConstraint(
                stoichiometries=self.objective_target,
                lower_value=lower_value,
                upper_value=upper_value,
            )
        ]
        maxz_model.extra_linear_constraints += [
            ExtraLinearConstraint(
                stoichiometries={f"{Z_VAR_PREFIX}{reac_id}": 1.0},
                upper_value=0.0,
            )
            for (reac_id, reac_data) in self.original_cobrak_model.reactions.items()
            if (reac_data.dG0 is not None) and (reac_id in deactivated_reactions)
        ]
        eligible_z_sum_objective = {
            f"{Z_VAR_PREFIX}{reac_id}": 1.0
            for (reac_id, reac_data) in self.original_cobrak_model.reactions.items()
            if (reac_data.dG0 is not None)
            and (self.variability_data[reac_id][1] > 0.0)
            and (reac_id not in deactivated_reactions)
        }
        try:
            maxz_ectfba_dict = perform_lp_optimization(
                cobrak_model=maxz_model,
                objective_target=eligible_z_sum_objective,
                objective_sense=+1,
                with_enzyme_constraints=True,
                with_thermodynamic_constraints=True,
                with_loop_constraints=True,
                variability_dict=deepcopy(self.variability_data),
                ignored_reacs=deactivated_reactions,
                solver=self.lp_solver,
                correction_config=self.correction_config,
            )
        except (ApplicationError, AttributeError, ValueError):
            maxz_ectfba_dict = {ALL_OK_KEY: False}
        if maxz_ectfba_dict[ALL_OK_KEY]:
            used_maxz_tfba_dict: dict[str, float] = {}
            for var_id in maxz_ectfba_dict:
                if var_id not in self.original_cobrak_model.reactions:
                    continue
                reaction = self.original_cobrak_model.reactions[var_id]
                if (
                    (reaction.dG0 is None) and (var_id not in deactivated_reactions)
                ) or (
                    (reaction.dG0 is not None)
                    and (maxz_ectfba_dict[f"{Z_VAR_PREFIX}{var_id}"] > 0.0)
                ):
                    used_maxz_tfba_dict[var_id] = 1.0
                else:
                    used_maxz_tfba_dict[var_id] = 0.0

            used_maxz_tfba_dict[ALL_OK_KEY] = True

            if used_maxz_tfba_dict[ALL_OK_KEY]:
                try:
                    second_nlp_dict = (
                        perform_nlp_irreversible_optimization_with_active_reacs_only(
                            cobrak_model=self.original_cobrak_model,
                            objective_target=self.objective_target,
                            objective_sense=self.objective_sense,
                            optimization_dict=deepcopy(used_maxz_tfba_dict),
                            variability_dict=deepcopy(self.variability_data),
                            with_kappa=self.with_kappa,
                            with_gamma=self.with_gamma,
                            with_iota=self.with_iota,
                            with_alpha=self.with_alpha,
                            solver=self.nlp_solver,
                            correction_config=self.correction_config,
                        )
                    )
                    if second_nlp_dict[ALL_OK_KEY] and (
                        abs(second_nlp_dict[OBJECTIVE_VAR_NAME]) > self.min_abs_objvalue
                    ):
                        nlp_results.append(second_nlp_dict)
                except (ApplicationError, AttributeError, ValueError):
                    pass

        ####
        try:
            minz_ectfba_dict = perform_lp_optimization(
                cobrak_model=maxz_model,
                objective_target=eligible_z_sum_objective,
                objective_sense=-1,
                with_enzyme_constraints=True,
                with_thermodynamic_constraints=True,
                with_loop_constraints=True,
                variability_dict=deepcopy(self.variability_data),
                ignored_reacs=deactivated_reactions,
                solver=self.lp_solver,
                correction_config=self.correction_config,
            )
        except (ApplicationError, AttributeError, ValueError):
            minz_ectfba_dict = {ALL_OK_KEY: False}
        if minz_ectfba_dict[ALL_OK_KEY]:
            used_minz_tfba_dict: dict[str, float] = {}
            for var_id in minz_ectfba_dict:
                if var_id not in self.original_cobrak_model.reactions:
                    continue
                reaction = self.original_cobrak_model.reactions[var_id]
                if (
                    (reaction.dG0 is None) and (var_id not in deactivated_reactions)
                ) or (
                    (reaction.dG0 is not None)
                    and (minz_ectfba_dict[f"{Z_VAR_PREFIX}{var_id}"] > 0.0)
                ):
                    used_minz_tfba_dict[var_id] = 1.0
                else:
                    used_minz_tfba_dict[var_id] = 0.0

            used_minz_tfba_dict[ALL_OK_KEY] = True

            if used_minz_tfba_dict[ALL_OK_KEY]:
                try:
                    third_nlp_dict = (
                        perform_nlp_irreversible_optimization_with_active_reacs_only(
                            cobrak_model=self.original_cobrak_model,
                            objective_target=self.objective_target,
                            objective_sense=self.objective_sense,
                            optimization_dict=deepcopy(used_minz_tfba_dict),
                            variability_dict=deepcopy(self.variability_data),
                            with_kappa=self.with_kappa,
                            with_gamma=self.with_gamma,
                            with_iota=self.with_iota,
                            with_alpha=self.with_alpha,
                            solver=self.nlp_solver,
                            correction_config=self.correction_config,
                        )
                    )
                    if third_nlp_dict[ALL_OK_KEY] and (
                        abs(third_nlp_dict[OBJECTIVE_VAR_NAME]) > self.min_abs_objvalue
                    ):
                        nlp_results.append(third_nlp_dict)
                except (ApplicationError, AttributeError, ValueError):
                    pass
        ####

        output: list[tuple[float, list[float | int]]] = [(1_000_000, [])]
        for nlp_result in nlp_results:
            objvalues = [nlp_result[OBJECTIVE_VAR_NAME] for nlp_result in nlp_results]
            if is_objsense_maximization(self.objective_sense):
                opt_idx = objvalues.index(max(objvalues))
            else:
                opt_idx = objvalues.index(min(objvalues))
            opt_nlp_dict = nlp_results[opt_idx]

            objective_value = opt_nlp_dict[OBJECTIVE_VAR_NAME]

            if self.temp_directory_name:
                filename = f"{self.temp_directory_name}{objective_value}{time()}{randint(0, 1_000_000_000)}.json"
                json_write(filename, opt_nlp_dict)

            if is_objsense_maximization(self.objective_sense):
                objective_value *= -1

            print("No error, objective value is:", objective_value)

            active_nlp_x: list[float | int] = [
                0 for _ in range(len(list(self.idx_to_reac_ids.keys())))
            ]
            for couple_idx, reac_ids in self.idx_to_reac_ids.items():
                reac_id = reac_ids[0]
                if reac_id not in opt_nlp_dict or opt_nlp_dict[reac_id] < 1e-11:
                    set_value = 0
                else:
                    set_value = 1
                active_nlp_x[couple_idx] = set_value
            output.append((objective_value, active_nlp_x))

        return output

    def optimize(self) -> dict[float, list[dict[str, float]]]:
        temp_directory = TemporaryDirectory()
        self.temp_directory_name = standardize_folder(temp_directory.name)

        match self.algorithm:
            case "pso":
                evolution = COBRAKPAPSO(
                    fitness_function=self.fitness,
                    xs_dim=self.dim,
                    extra_xs=self.initial_xs_list,
                    gen=self.num_gens,
                    objvalue_json_path=self.objvalue_json_path,
                    max_rounds_same_objvalue=self.max_rounds_same_objvalue,
                    pop_size=self.pop_size,
                )
            case "genetic":
                evolution = COBRAKGENETIC(
                    fitness_function=self.fitness,
                    xs_dim=self.dim,
                    extra_xs=self.initial_xs_list,
                    gen=self.num_gens,
                    objvalue_json_path=self.objvalue_json_path,
                    max_rounds_same_objvalue=self.max_rounds_same_objvalue,
                    pop_size=self.pop_size,
                )
            case _:
                print(
                    f"ERROR: Evolution algorithm {self.algorithm} does not exist! Use either 'pso' or 'genetic'."
                )
                raise ValueError
        evolution.run()

        result_dict: dict[float, list[dict[str, float]]] = {}
        for json_filename in get_files(self.temp_directory_name):
            json_data = json_load(f"{self.temp_directory_name}{json_filename}", Any)
            objective_value = json_data[OBJECTIVE_VAR_NAME]
            if objective_value not in result_dict:
                result_dict[objective_value] = []
            result_dict[objective_value].append(deepcopy(json_data))

        temp_directory.cleanup()

        return {
            key: result_dict[key] for key in sorted(result_dict.keys(), reverse=True)
        }


from pyomo.environ import Binary, Constraint, Reals, Var

from .constants import BIG_M
from .lps import add_statuses_to_optimziation_dict, get_lp_from_cobrak_model
from .pyomo_functionality import (
    add_objective_to_model,
    get_solver,
)
from .utilities import (
    add_objective_value_as_extra_linear_constraint,
    apply_variability_dict,
    delete_orphaned_metabolites_and_enzymes,
    get_pyomo_solution_as_dict,
)


def _postprocess_batch(
    reac_couples: list[str],
    target_couples: list[tuple[str, list[str]]],
    active_reacs: list[str],
    cobrak_model: Model,
    objective_target: str | dict[str, float],
    objective_sense: int,
    variability_data: dict[str, tuple[float, float]],
    pyomo_lp_solver,
    with_kappa: bool = True,
    with_gamma: bool = True,
    with_iota: bool = False,
    with_alpha: bool = False,
    lp_solver: Solver = SCIP,
    nlp_solver: Solver = IPOPT,
    verbose: bool = False,
    correction_config: CorrectionConfig = CorrectionConfig(),
    onlytested: str = "",
):
    original_var_data = deepcopy(variability_data)
    feasible_switches = []
    for target_type, target_couple, max_allowed_changes in target_couples:
        if onlytested and (not any(onlytested in reac_id for reac_id in target_couple)):
            continue

        for at_maximum in (False,):
            target_cobrak_model = deepcopy(cobrak_model)
            variability_data = deepcopy(original_var_data)
            if target_type == "deac":
                print(
                    f"DEACTIVATING {target_couple} {at_maximum}, max changes: {max_allowed_changes}"
                )
                for reac_id in target_couple:
                    del target_cobrak_model.reactions[reac_id]
                target_cobrak_model = delete_orphaned_metabolites_and_enzymes(
                    target_cobrak_model
                )
            else:
                print(
                    f"ACTIVATING {target_couple} {at_maximum}, max changes: {max_allowed_changes}"
                )
                variability_data[target_couple[0]] = (
                    1e-5,
                    variability_data[target_couple[0]][1],
                )

            baselp = get_lp_from_cobrak_model(
                cobrak_model=target_cobrak_model,
                with_enzyme_constraints=True,
                with_thermodynamic_constraints=True,
                with_loop_constraints=False,
                with_flux_sum_var=False,
                add_extra_linear_constraints=True,
                correction_config=correction_config,
            )
            baselp = apply_variability_dict(
                baselp,
                target_cobrak_model,
                variability_data,
                error_scenario=correction_config.error_scenario,
            )

            active_z_var_changes_sum = 0.0
            for active_reac in active_reacs:
                if (
                    active_reac in target_couple
                    or target_cobrak_model.reactions[active_reac].dG0 is None
                ):
                    continue
                z_var_id = f"{Z_VAR_PREFIX}{active_reac}"
                z_var_change_id = f"{z_var_id}_change"
                setattr(baselp, z_var_change_id, Var(within=Binary))
                setattr(
                    baselp,
                    f"fix_{z_var_id}",
                    Constraint(
                        expr=getattr(baselp, z_var_id)
                        == 1.0 - getattr(baselp, z_var_change_id)
                    ),
                )
                active_z_var_changes_sum += getattr(baselp, z_var_change_id)
            max_z_var_changes_id = "max_active_z_var_changes"
            setattr(
                baselp,
                max_z_var_changes_id,
                Var(within=Reals),
            )
            getattr(baselp, max_z_var_changes_id).lb = 0.0
            getattr(baselp, max_z_var_changes_id).ub = max_allowed_changes
            setattr(
                baselp,
                "active_original_z_vars_sum",
                Constraint(
                    expr=active_z_var_changes_sum
                    <= getattr(baselp, max_z_var_changes_id)
                ),
            )

            baselp = add_objective_to_model(
                baselp,
                objective_target,
                objective_sense,
                "ORIGINAL_OBJ",
                "ORIGINAL_OBJ_VAR",
            )
            results = pyomo_lp_solver.solve(
                baselp, tee=verbose, **lp_solver.solve_extra_options
            )
            lp_resultdict = add_statuses_to_optimziation_dict(
                get_pyomo_solution_as_dict(baselp), results
            )
            if not lp_resultdict[ALL_OK_KEY]:
                continue
            new_active_reacs = get_active_reacs_from_optimization_dict(
                target_cobrak_model, lp_resultdict
            )
            getattr(baselp, "ORIGINAL_OBJ").deactivate()
            if at_maximum:
                getattr(baselp, "ORIGINAL_OBJ_VAR").lb = (
                    getattr(baselp, "ORIGINAL_OBJ_VAR").value - 1e-6
                )

            for active_reac in new_active_reacs:
                if (
                    active_reac in target_couple
                    or target_cobrak_model.reactions[active_reac].dG0 is None
                ):
                    continue
                if active_reac in active_reacs:
                    continue
                getattr(baselp, f"{Z_VAR_PREFIX}{active_reac}").lb = 1.0

            EXTRAZ_PREFIX = "extra_z_var_"
            extraz_names: list[str] = []
            for reac_couple in reac_couples:
                if target_couple == reac_couple:
                    continue
                if reac_couple[0] in active_reacs:
                    continue
                if all(
                    target_cobrak_model.reactions[reac_id].enzyme_reaction_data is None
                    and target_cobrak_model.reactions[reac_id].dG0 is None
                    for reac_id in reac_couple
                ):
                    continue
                if variability_data[reac_couple[0]][1] == 0.0:
                    continue

                extraz_names.append(f"{EXTRAZ_PREFIX}_of_couple_of_{reac_couple[0]}")
                setattr(
                    baselp,
                    extraz_names[-1],
                    Var(
                        within=Binary,
                    ),
                )
                setattr(
                    baselp,
                    f"{EXTRAZ_PREFIX}_constraint_of_couple_of_{reac_couple[0]}",
                    Constraint(
                        expr=getattr(baselp, reac_couple[0])
                        <= BIG_M * getattr(baselp, extraz_names[-1])
                    ),
                )
                couple_reacs_with_dG0 = [
                    reac_id
                    for reac_id in target_cobrak_model.reactions
                    if target_cobrak_model.reactions[reac_id].dG0 is not None
                ]
                if len(couple_reacs_with_dG0) == 0:
                    continue
                setattr(
                    baselp,
                    f"{EXTRAZ_PREFIX}_z_var_constraint_of_couple_of_{reac_couple[0]}",
                    Constraint(
                        expr=getattr(
                            baselp, f"{Z_VAR_PREFIX}{couple_reacs_with_dG0[0]}"
                        )
                        >= getattr(baselp, extraz_names[-1])
                    ),
                )

            extraz_objective = dict.fromkeys(extraz_names, 1.0)

            for extraz_objective_sense in (-1, +1):
                baselp = add_objective_to_model(
                    baselp,
                    extraz_objective,
                    extraz_objective_sense,
                    f"postprocess_obj_{extraz_objective_sense}",
                    f"postprocess_obj_var_{extraz_objective_sense}",
                )
                getattr(
                    baselp, f"postprocess_obj_{extraz_objective_sense}"
                ).deactivate()

            for extraz_objective_sense in (-1, +1):
                getattr(baselp, f"postprocess_obj_{extraz_objective_sense}").activate()
                try:
                    results = pyomo_lp_solver.solve(
                        baselp, tee=verbose, **lp_solver.solve_extra_options
                    )
                except (ApplicationError, AttributeError, ValueError):
                    lp_resultdict = {ALL_OK_KEY: False}
                try:
                    lp_resultdict = add_statuses_to_optimziation_dict(
                        get_pyomo_solution_as_dict(baselp), results
                    )
                except (ApplicationError, AttributeError, ValueError):
                    lp_resultdict = {ALL_OK_KEY: False}
                getattr(
                    baselp, f"postprocess_obj_{extraz_objective_sense}"
                ).deactivate()
                if not lp_resultdict[ALL_OK_KEY]:
                    continue

                try:
                    nlp_resultdict = (
                        perform_nlp_irreversible_optimization_with_active_reacs_only(
                            cobrak_model=target_cobrak_model,
                            objective_target=objective_target,
                            objective_sense=objective_sense,
                            optimization_dict=deepcopy(lp_resultdict),
                            variability_dict=deepcopy(variability_data),
                            with_kappa=with_kappa,
                            with_gamma=with_gamma,
                            with_iota=with_iota,
                            with_alpha=with_alpha,
                            solver=nlp_solver,
                            correction_config=correction_config,
                        )
                    )
                    if nlp_resultdict[ALL_OK_KEY]:
                        feasible_switches.append(
                            [
                                target_couple,
                                tuple(
                                    extraz_var
                                    for extraz_var, value in lp_resultdict.items()
                                    if extraz_var.startswith(EXTRAZ_PREFIX)
                                    and value > 0.1
                                ),
                                nlp_resultdict,
                                f"{target_type}{at_maximum}",
                                max_allowed_changes,
                            ],
                        )
                except (ApplicationError, AttributeError, ValueError):
                    pass
    return feasible_switches


def postprocess(
    cobrak_model: Model,
    opt_dict: dict[str, float],
    objective_target: str | dict[str, float],
    objective_sense: int,
    variability_data: dict[str, tuple[float, float]],
    with_kappa: bool = True,
    with_gamma: bool = True,
    with_iota: bool = False,
    with_alpha: bool = False,
    lp_solver: Solver = SCIP,
    nlp_solver: Solver = IPOPT,
    verbose: bool = False,
    correction_config: CorrectionConfig = CorrectionConfig(),
    onlytested: str = "",
) -> tuple[float, list[float | int]]:
    if variability_data == {}:
        variability_data = perform_lp_variability_analysis(
            cobrak_model=cobrak_model,
            with_enzyme_constraints=True,
            with_thermodynamic_constraints=True,
            active_reactions=[],
            solver=lp_solver,
        )
    else:
        variability_data = deepcopy(variability_data)

    pyomo_lp_solver = get_solver(
        lp_solver.name,
        lp_solver.solver_options,
        lp_solver.solver_attrs,
    )

    cobrak_model = deepcopy(cobrak_model)
    if type(objective_target) is str:
        objective_target = {objective_target: 1.0}
    obj_value = 0.0
    for obj_target_id, obj_target_multiplier in objective_target.items():
        obj_value += opt_dict[obj_target_id] * obj_target_multiplier
    epsilon = 1e-6 if is_objsense_maximization(objective_sense) else -1e-6
    cobrak_model = add_objective_value_as_extra_linear_constraint(
        cobrak_model,
        obj_value + epsilon,
        objective_target,
        objective_sense,
    )

    reac_couples = get_stoichiometrically_coupled_reactions(cobrak_model)
    active_reacs = [
        active_reac
        for active_reac in get_active_reacs_from_optimization_dict(
            cobrak_model, opt_dict
        )
        if (active_reac in [reac_ids[0] for reac_ids in reac_couples])
        and (active_reac not in objective_target)  # and opt_dict[active_reac] > 1e-8
    ]
    active_reac_couples: list[list[str]] = [
        reac_couple
        for reac_couple in reac_couples
        if reac_couple[0] in active_reacs and variability_data[reac_couple[0]][0] == 0.0
    ]
    inactive_reac_couples: list[list[str]] = [
        reac_couple
        for reac_couple in reac_couples
        if reac_couple[0] not in active_reacs
        and variability_data[reac_couple[0]][1] > 0.0
    ]
    targets = []
    for max_target_num in (0, 5):
        targets += [("deac", x, max_target_num) for x in active_reac_couples]
        targets += [("ac", x, max_target_num) for x in inactive_reac_couples]
    all_feasible_switches_metalist = Parallel(n_jobs=-1, verbose=10)(
        delayed(_postprocess_batch)(
            reac_couples,
            targets_batch,
            active_reacs,
            cobrak_model,
            objective_target,
            objective_sense,
            variability_data,
            pyomo_lp_solver,
            with_kappa,
            with_gamma,
            with_iota,
            with_alpha,
            lp_solver,
            nlp_solver,
            verbose,
            correction_config,
            onlytested,
        )
        for targets_batch in split_list(targets, cpu_count())
    )
    all_feasible_switches = []
    for sublist in all_feasible_switches_metalist:
        all_feasible_switches.extend(sublist)

    if len(all_feasible_switches) > 0:
        best_result = all_feasible_switches[0][2]
        for result in [x[2] for x in all_feasible_switches[1:]]:
            if is_objsense_maximization(objective_sense):
                if result[OBJECTIVE_VAR_NAME] > best_result[OBJECTIVE_VAR_NAME]:
                    best_result = result
            else:
                if result[OBJECTIVE_VAR_NAME] < best_result[OBJECTIVE_VAR_NAME]:
                    best_result = result
    else:
        best_result = {}

    return all_feasible_switches, best_result
    ####


def _sampling_routine(
    cobrak_model: Model,
    objective_target: str,
    objective_sense: int,
    variability_dict: dict[str, tuple[float, float]],
    with_kappa: bool,
    with_gamma: bool,
    with_iota: bool,
    with_alpha: bool,
    deactivated_reaction_lists: list[list[str]],
    lp_solver: Solver,
    nlp_solver: Solver,
    correction_config: CorrectionConfig,
    min_abs_objvalue: float,
) -> list[dict[str, float]]:
    working_dicts: list[dict[str, float]] = []
    for deactivated_reaction_set in deactivated_reaction_lists:
        try:
            ectfba_dict = perform_lp_optimization(
                cobrak_model=cobrak_model,
                objective_target=objective_target,
                objective_sense=objective_sense,
                with_enzyme_constraints=True,
                with_thermodynamic_constraints=True,
                with_loop_constraints=False,
                variability_dict=variability_dict,
                ignored_reacs=deactivated_reaction_set,
                solver=lp_solver,
                correction_config=correction_config,
            )
        except (ApplicationError, AttributeError, ValueError):
            continue

        if not ectfba_dict[ALL_OK_KEY]:
            continue

        active_ectfba_reacs = get_active_reacs_from_optimization_dict(
            cobrak_model, ectfba_dict
        )
        missing_errortarget = False
        for errortarget in correction_config.error_scenario:
            if errortarget not in cobrak_model.reactions:
                continue
            if errortarget not in active_ectfba_reacs:
                missing_errortarget = True
        if missing_errortarget:
            continue

        ###################################################
        # 3. NLP for initial set of rmax guesses
        try:
            nlp_result = perform_nlp_irreversible_optimization_with_active_reacs_only(
                cobrak_model=cobrak_model,
                objective_target=objective_target,
                objective_sense=objective_sense,
                optimization_dict=ectfba_dict,
                variability_dict=variability_dict,
                with_kappa=with_kappa,
                with_gamma=with_gamma,
                with_iota=with_iota,
                with_alpha=with_alpha,
                verbose=False,
                solver=nlp_solver,
                correction_config=correction_config,
            )
        except (ApplicationError, AttributeError, ValueError):
            continue
        if not nlp_result[ALL_OK_KEY]:
            continue
        if abs(nlp_result[OBJECTIVE_VAR_NAME]) < min_abs_objvalue:
            continue
        if nlp_result[ALL_OK_KEY]:
            working_dicts.append(deepcopy(nlp_result))
            print("Working sampling result:", nlp_result[OBJECTIVE_VAR_NAME])
    return working_dicts


def perform_nlp_evolutionary_optimization(
    cobrak_model: Model,
    objective_target: str | dict[str, float],
    objective_sense: int,
    variability_dict: dict[str, tuple[float, float]] = {},
    with_kappa: bool = True,
    with_gamma: bool = True,
    with_iota: bool = False,
    with_alpha: bool = False,
    sampling_wished_num_feasible_starts: int = 3,
    sampling_max_metarounds: int = 3,
    sampling_rounds_per_metaround: int = 2,
    sampling_max_deactivated_reactions: int = 5,
    sampling_always_deactivated_reactions: list[str] = [],
    evolution_num_gens: int = 5,
    algorithm: Literal["pso", "genetic"] = "genetic",
    lp_solver: Solver = SCIP,
    nlp_solver: Solver = IPOPT,
    objvalue_json_path: str = "",
    max_rounds_same_objvalue: float = float("inf"),
    correction_config: CorrectionConfig = CorrectionConfig(),
    min_abs_objvalue: float = 1e-13,
    pop_size: int | None = None,
    working_results: list[dict[str, float]] = [],
) -> dict[float, list[dict[str, float]]]:
    if variability_dict == {}:
        variability_dict = perform_lp_variability_analysis(
            cobrak_model=cobrak_model,
            with_enzyme_constraints=True,
            with_thermodynamic_constraints=True,
            active_reactions=[],
            solver=lp_solver,
        )
    else:
        variability_dict = deepcopy(variability_dict)

    # Initial sampling
    if isinstance(objective_target, str):
        objective_target = {objective_target: 1.0}
    objective_target_ids = list(objective_target.keys())  # type: ignore

    deactivatable_reactions = [
        var_id
        for var_id in variability_dict
        if (var_id in cobrak_model.reactions)
        and (variability_dict[var_id][0] == 0.0)
        and (var_id not in objective_target_ids)
        and (var_id not in sampling_always_deactivated_reactions)
        and (var_id not in correction_config.error_scenario)
    ]
    distinct_feasible_start_solutions: dict[tuple[str, ...], dict[str, float]] = {}
    for current_round in range(sampling_max_metarounds):
        # Get deactivated reaction lists
        all_deactivated_reaction_lists = [
            [
                sample(
                    deactivatable_reactions,
                    randint(1, sampling_max_deactivated_reactions + 1),
                )
                + sampling_always_deactivated_reactions
                for _ in range(sampling_rounds_per_metaround)
            ]
            for _ in range(cpu_count())
        ]
        if current_round == 0:
            all_deactivated_reaction_lists[0][0] = deepcopy(
                sampling_always_deactivated_reactions
            )

        # run sampling
        results = Parallel(n_jobs=-1, verbose=10)(
            delayed(_sampling_routine)(
                cobrak_model,
                objective_target,
                objective_sense,
                variability_dict,
                with_kappa,
                with_gamma,
                with_iota,
                with_alpha,
                deactivated_reaction_lists,
                lp_solver,
                nlp_solver,
                correction_config,
                min_abs_objvalue,
            )
            for deactivated_reaction_lists in all_deactivated_reaction_lists
        )
        if len(working_results) > 0:
            results.append(working_results)
        best_result = (
            -float("inf") if is_objsense_maximization(objective_sense) else float("inf")
        )
        for result in results:
            for tfba_dict in result:
                active_reacs_tuple = tuple(
                    sorted(
                        get_active_reacs_from_optimization_dict(cobrak_model, tfba_dict)
                    )
                )
                distinct_feasible_start_solutions[active_reacs_tuple] = deepcopy(
                    tfba_dict
                )
                if is_objsense_maximization(objective_sense):
                    best_result = max(tfba_dict[OBJECTIVE_VAR_NAME], best_result)
                else:
                    best_result = min(tfba_dict[OBJECTIVE_VAR_NAME], best_result)

        if (
            len(distinct_feasible_start_solutions.keys())
            >= sampling_wished_num_feasible_starts
        ):
            break

    print(best_result)

    if len(distinct_feasible_start_solutions.keys()) == 0:
        print(
            "ERROR: No feasible sampling solution found! Check feasibility of problem and/or adjust sampling settings."
        )
        raise ValueError
    if (
        len(distinct_feasible_start_solutions.keys())
        < sampling_wished_num_feasible_starts
    ):
        print("INFO: Fewer feasible sampling solutions found than wished.")

    # Evolutionary algorithm
    problem = COBRAKProblem(
        cobrak_model=cobrak_model,
        objective_target=objective_target,  # type: ignore
        objective_sense=objective_sense,
        variability_dict=variability_dict,
        nlp_dict_list=list(distinct_feasible_start_solutions.values()),
        best_value=best_result,
        with_kappa=with_kappa,
        with_gamma=with_gamma,
        with_iota=with_iota,
        with_alpha=with_alpha,
        num_gens=evolution_num_gens,
        algorithm=algorithm,
        lp_solver=lp_solver,
        nlp_solver=nlp_solver,
        objvalue_json_path=objvalue_json_path,
        max_rounds_same_objvalue=max_rounds_same_objvalue,
        correction_config=correction_config,
        min_abs_objvalue=min_abs_objvalue,
        pop_size=pop_size,
    )

    return problem.optimize()



import operator
from collections.abc import Callable
from copy import deepcopy
from os import cpu_count
from random import choice, randint, uniform
from time import time

import numpy as np
from joblib import Parallel, delayed

from .io import json_write
from .utilities import count_last_equal_elements, last_n_elements_equal


class COBRAKGENETIC:
    def __init__(
        self,
        fitness_function: Callable[
            [list[float | int]], tuple[float, list[float | int]]
        ],
        xs_dim: int,
        gen: int,
        extra_xs: list[list[int]] = [],
        seed: int | None = None,
        objvalue_json_path: str = "",
        max_rounds_same_objvalue: float = float("inf"),
        pop_size: int | None = None,
    ) -> None:
        # Parameters
        self.fitness_function = fitness_function
        self.xs_dim = xs_dim
        self.gen = gen
        self.seed = seed
        if seed is not None:
            np.random.seed(seed)

        # Initialization of random particles
        cpu_count_value = cpu_count() if pop_size is None else pop_size
        if cpu_count_value is None:
            self.cpu_count = 1
        else:
            self.cpu_count = cpu_count_value
        self.init_xs = [
            [randint(0, 1) for _ in range(xs_dim)]
            for _ in range(self.cpu_count - len(extra_xs))
        ]

        # Addition of user-defined extra particles
        if extra_xs != []:
            self.init_xs.extend(extra_xs)

        self.tested_xs: dict[tuple[int, ...], float] = {}
        self.all_xs: dict[tuple[int, ...], float] = {}
        self.objvalue_json_path = objvalue_json_path
        self.objvalue_json_data: dict[float, list[float]] = {}
        self.max_rounds_same_objvalue = max_rounds_same_objvalue

    def _get_sorted_list_from_tested_xs(self) -> list[tuple[float, tuple[int, ...]]]:
        return sorted(
            [(fitness, x) for (x, fitness) in self.tested_xs.items()],
            key=operator.itemgetter(0),
        )

    def run(self) -> tuple[float, tuple[int, ...]]:
        init_fitnesses = Parallel(n_jobs=-1)(
            delayed(self.fitness_function)(x) for x in self.init_xs
        )
        if init_fitnesses is not None:
            self.tested_xs = {}
            for init_fitness in init_fitnesses:
                for fitness, xs in init_fitness:
                    self.tested_xs[tuple(xs)] = fitness
        else:
            print("ERROR: Something went wrong during initialization")
            raise ValueError

        if self.objvalue_json_path:
            start_time = time()
            self.objvalue_json_data[0.0] = sorted(self.tested_xs.values())
            json_write(self.objvalue_json_path, self.objvalue_json_data)

        # Actual algorithm
        max_objvalues = []
        for _ in range(self.gen):
            max_objvalues.append(max(self.tested_xs.values()))
            if last_n_elements_equal(max_objvalues, self.max_rounds_same_objvalue):  # type: ignore
                break

            xs_list = self._get_sorted_list_from_tested_xs()
            xs_list = [xs for xs in xs_list if len(xs) > 0]

            chosen_xs: list[tuple[int, ...]] = []
            # Choose some of the top 3
            for _ in range(self.cpu_count // 4):
                chosen_xs.append(choice([x[1] for x in xs_list if len(x[1]) > 0][:3]))

            # Choose some of the 25% best
            for _ in range(self.cpu_count // 2):
                chosen_xs.append(
                    choice([x[1] for x in xs_list][: round(len(xs_list) * 0.25)])
                )

            # Choose some other 75% worst
            for _ in range(self.cpu_count // 4):
                addlength = 0
                while not addlength:
                    added_xs = deepcopy(
                        choice([x[1] for x in xs_list][round(len(xs_list) * 0.25) :])
                    )
                    addlength = len(added_xs)
                chosen_xs.append(added_xs)

            # Random crossovers
            for _ in range(round(len(chosen_xs) * 0.2)):
                target = randint(0, len(chosen_xs) - 1)
                source = randint(0, len(chosen_xs) - 1)
                cut = randint(0, self.xs_dim - 1)
                chosen_xs[target] = deepcopy(chosen_xs[target][:cut]) + deepcopy(
                    chosen_xs[source][cut:]
                )

            # Test Xs in parallel
            results = Parallel(n_jobs=-1, verbose=10)(
                delayed(self.update_particle)(
                    chosen_x,
                    count_last_equal_elements(max_objvalues),
                )
                for chosen_x in chosen_xs
            )

            if results is None:
                print("ERROR: Something went wrong during fitness calculations")
                raise ValueError

            # Unpack results
            for fitnesses_and_active_xs, mutated_x in results:
                for fitness, active_x in fitnesses_and_active_xs:
                    if active_x is None or active_x == []:
                        continue
                    self.tested_xs[tuple(active_x)] = fitness
                    self.all_xs[tuple(active_x)] = fitness
                self.all_xs[tuple(mutated_x)] = max(
                    fitness for (fitness, _) in fitnesses_and_active_xs
                )

            if self.objvalue_json_path:
                self.objvalue_json_data[time() - start_time] = sorted(
                    self.tested_xs.values()
                )
                json_write(self.objvalue_json_path, self.objvalue_json_data)

        best_f_and_x = self._get_sorted_list_from_tested_xs()[0]
        return best_f_and_x[0], best_f_and_x[1]

    def update_particle(
        self,
        chosen_x: list[int],
        num_rounds_without_best_change: int,
    ) -> tuple[float, list[int], list[int]]:
        if not len(chosen_x):
            return [[1_000_000, []]], []

        min_change_p = 0.1 * 0.95**num_rounds_without_best_change
        max_change_p = 0.1 * 1.05**num_rounds_without_best_change
        change_p = uniform(min_change_p, max_change_p)
        change_p = max(0.001, change_p)
        change_p = min(0.999, change_p)

        mutation_tries = 0
        while True:
            mutated_x: list[float] = []
            match randint(0, 2):
                case 0:  # Extend
                    for x in chosen_x:
                        if x == 1:
                            mutated_x.append(1)
                            continue
                        if uniform(0.0, 1.0) < change_p:
                            mutated_x.append(1)
                        else:
                            mutated_x.append(x)
                case 1:  # Decrease
                    for x in chosen_x:
                        if x == 0:
                            mutated_x.append(0)
                            continue
                        if uniform(0.0, 1.0) < change_p:
                            mutated_x.append(0)
                        else:
                            mutated_x.append(x)
                case 2:  # Extend and decrease
                    for x in chosen_x:
                        if x == 1:
                            if uniform(0.0, 1.0) < change_p:
                                mutated_x.append(0)
                            else:
                                mutated_x.append(x)
                        else:
                            if uniform(0.0, 1.0) < change_p:
                                mutated_x.append(1)
                            else:
                                mutated_x.append(x)
                # case 3:  # Random
                #    mutated_x = [randint(0, 1) for _ in range(len(chosen_x))]

            if mutation_tries > 250:
                return [[1_000_000, []]], []
            if tuple(mutated_x) in self.all_xs:
                mutation_tries += 1
            else:
                break

        # Evaluate new position
        fitnesses_and_active_xs = self.fitness_function(mutated_x)

        return fitnesses_and_active_xs, mutated_x



# IMPORTS SECION #
from numpy import exp

from cobrak.constants import (
    DF_VAR_PREFIX,
    GAMMA_VAR_PREFIX,
    KAPPA_VAR_PREFIX,
    LNCONC_VAR_PREFIX,
    STANDARD_R,
    STANDARD_T,
)
from cobrak.dataclasses import Model, Reaction
from cobrak.utilities import (
    delete_orphaned_metabolites_and_enzymes,
    get_unoptimized_reactions_in_nlp_solution,
    have_all_unignored_km,
)


# "PRIVATE" FUNCTIONS SECTION #
def _get_numbersafe_id(met_id: str) -> str:
    match met_id.startswith(("0", "1", "2", "3", "4", "5", "6", "7", "8", "9")):
        case True:
            return f"x{met_id}"
        case False:
            return met_id


def _get_reaction_string_of_cobrak_reaction(
    cobrak_model: Model,
    reac_id: str,
    cobrak_reaction: Reaction,
    e_conc: float,
    met_concs: dict[str, float],
    reac_flux: float,
    nlp_results: dict[str, float],
    kinetic_ignored_metabolites: list[str],
    unoptimized_reactions: dict[str, tuple[float, float]],
) -> str:
    reac_id = _get_numbersafe_id(reac_id)

    has_vplus = (
        (cobrak_reaction.enzyme_reaction_data is not None)
        and (cobrak_reaction.enzyme_reaction_data.k_cat is not None)
        and (cobrak_reaction.enzyme_reaction_data.identifiers != [""])
    )
    has_kappa = has_vplus and have_all_unignored_km(
        cobrak_reaction, kinetic_ignored_metabolites
    )
    has_gamma = cobrak_reaction.dG0 is not None

    # Reaction stoichiometry
    pow_stoich_substrate_strings = []
    pow_stoich_d_km_substrate_strings = []
    pow_stoich_product_strings = []
    pow_stoich_d_km_product_strings = []
    stoich_times_substrate_strings = []
    stoich_times_product_strings = []
    substrates = []
    products = []
    for unsafe_met_id, stoichiometry in cobrak_reaction.stoichiometries.items():
        met_id = _get_numbersafe_id(unsafe_met_id)
        stoich_times_met_string = str(abs(stoichiometry)) + " " + met_id
        pow_stoich_met_string = (
            "(" + met_id + "_molar" + "^" + str(abs(stoichiometry)) + ")"
        )
        pow_stoich_d_km_met_string = (
            "(("
            + met_id
            + "_molar"
            + f" / km_{met_id}_{reac_id}"
            + ")"
            + "^"
            + str(abs(stoichiometry))
            + ")"
        )
        if met_id in kinetic_ignored_metabolites:
            pow_stoich_d_km_met_string = "1.0"
        if stoichiometry < 0:
            stoich_times_substrate_strings.append(stoich_times_met_string)
            pow_stoich_substrate_strings.append(pow_stoich_met_string)
            pow_stoich_d_km_substrate_strings.append(pow_stoich_d_km_met_string)
            substrates.append(met_id)
        else:
            stoich_times_product_strings.append(stoich_times_met_string)
            pow_stoich_product_strings.append(pow_stoich_met_string)
            pow_stoich_d_km_product_strings.append(pow_stoich_d_km_met_string)
            products.append(met_id)
    reac_string = f"{reac_id}: {' + '.join(stoich_times_substrate_strings)} -> {' + '.join(stoich_times_product_strings)};\n"

    kappa_substrates = "(" + " * ".join(pow_stoich_d_km_substrate_strings) + ")"
    gamma_substrates = "(" + " * ".join(pow_stoich_substrate_strings) + ")"
    kappa_products = "(" + " * ".join(pow_stoich_d_km_product_strings) + ")"
    gamma_products = "(" + " * ".join(pow_stoich_product_strings) + ")"

    if reac_id in unoptimized_reactions:
        nlp_flux = unoptimized_reactions[reac_id][0]
        real_flux = unoptimized_reactions[reac_id][1]
        e_conc *= nlp_flux / real_flux

    reac_variables = f"const E_conc_{reac_id} = {e_conc}\n"

    if (not has_vplus) or (not has_kappa and not has_gamma):
        substrate_mult_strlist: list[str] = []
        product_mult_strlist: list[str] = []

        total_mult = 1.0
        for met_id, stoichiometry in cobrak_reaction.stoichiometries.items():
            met_var_id = LNCONC_VAR_PREFIX + met_id
            met_var_id_2 = LNCONC_VAR_PREFIX + met_id[1:]
            if met_id in met_concs:
                used_conc = met_concs[met_id]
            elif met_var_id in nlp_results:
                used_conc = exp(nlp_results[met_var_id])
            elif met_var_id_2 in nlp_results:
                used_conc = exp(nlp_results[met_var_id_2])
            else:
                used_conc = exp(cobrak_model.metabolites[met_id].log_min_conc)

            total_mult *= used_conc**stoichiometry
            if stoichiometry < 0.0:
                substrate_mult_strlist.append(
                    f"({_get_numbersafe_id(met_id)}_molar ^ {abs(stoichiometry)})"
                )
            else:
                product_mult_strlist.append(
                    f"({_get_numbersafe_id(met_id)}_molar ^ {stoichiometry})"
                )

        k_eq = 1.1 * total_mult

        real_flux = 1 - (total_mult / k_eq)
        kappa_mult_str = (
            " * ".join(product_mult_strlist)
            + " / ("
            + " * ".join(substrate_mult_strlist)
            + ") / "
            + str(k_eq)
        )

        reac_assignments = f"{KAPPA_VAR_PREFIX}{reac_id} := {kappa_mult_str}\n"
        menten_multiplier = reac_flux / real_flux
        reac_kinetic = f"{reac_id} = {menten_multiplier} * (1 - {kappa_mult_str})\n"
    else:
        if cobrak_reaction.enzyme_reaction_data is None:
            raise ValueError

        flux_value = 1.0

        # V plus
        reac_variables += (
            f"const k_cat_{reac_id} = {cobrak_reaction.enzyme_reaction_data.k_cat}\n"
        )
        v_plus = f"E_conc_{reac_id} * k_cat_{reac_id}"
        reac_assignments = f"v_plus_{reac_id} := {v_plus}\n"
        flux_value *= cobrak_reaction.enzyme_reaction_data.k_cat * e_conc

        if has_kappa:
            kappa_substrates_value = 1.0
            kappa_products_value = 1.0
            for unsafe_met_id, km in cobrak_reaction.enzyme_reaction_data.k_ms.items():
                original_met_id = unsafe_met_id
                met_id = _get_numbersafe_id(unsafe_met_id)
                reac_variables += f"const km_{met_id}_{reac_id} = {km}\n"

                stoichiometry = cobrak_reaction.stoichiometries[original_met_id]
                kappa_mult = (
                    exp(nlp_results[LNCONC_VAR_PREFIX + original_met_id]) / km
                ) ** abs(stoichiometry)
                if stoichiometry < 0:
                    kappa_substrates_value *= kappa_mult
                else:
                    kappa_products_value *= kappa_mult
            flux_value *= kappa_substrates_value / (
                1 + kappa_substrates_value + kappa_products_value
            )

            kappa = f"{kappa_substrates} / (1 + {kappa_substrates} + {kappa_products})"
            reac_assignments += f"{KAPPA_VAR_PREFIX}{reac_id} := {kappa}\n"
        else:
            kappa = "1"

        if has_gamma:
            used_dG0 = cobrak_reaction.dG0
            if used_dG0 is None:
                raise ValueError

            k_eq = exp(-used_dG0 / (cobrak_model.R * cobrak_model.T))

            reac_variables += f"const dG0_{reac_id} = {used_dG0}\n"
            reac_variables += f"const k_eq_{reac_id} = {k_eq}\n"

            dg = f"(dG0_{reac_id} + R * T * log( ({gamma_products}) ) - R * T * log( ({gamma_substrates}) ))"
            gamma = f"(1 - ({gamma_products} / {gamma_substrates} / k_eq_{reac_id}))"
            reac_assignments += f"{DF_VAR_PREFIX}{reac_id} := -{dg}\n"
            reac_assignments += f"{GAMMA_VAR_PREFIX}{reac_id} := {gamma}\n"

            gamma_products_mult = 1.0
            gamma_substrates_mult = 1.0
            for met_id, stoichiometry in cobrak_reaction.stoichiometries.items():
                if stoichiometry < 0.0:
                    gamma_substrates_mult *= exp(
                        nlp_results[LNCONC_VAR_PREFIX + met_id]
                    ) ** abs(stoichiometry)
                else:
                    gamma_products_mult *= (
                        exp(nlp_results[LNCONC_VAR_PREFIX + met_id]) ** stoichiometry
                    )
            flux_value *= 1 - (gamma_products_mult / gamma_substrates_mult / k_eq)
        else:
            gamma = "1"

        reac_kinetic = f"{reac_id} = ({reac_flux / flux_value} * {v_plus})  *  ({kappa})  *  ({gamma});\n"

    reac_string += reac_kinetic + reac_variables + reac_assignments

    return reac_string


# "PUBLIC" FUNCTIONS SECTION #
def get_tellurium_string_from_cobrak_model(
    cobrak_model: Model,
    cell_density: float,
    e_concs: dict[str, float],
    met_concs: dict[str, float],
    nlp_results: dict[str, float],
) -> str:
    unoptimized_reactions = get_unoptimized_reactions_in_nlp_solution(
        cobrak_model,
        nlp_results,
    )
    tellurium_string = (
        "# General constants\n" + f"R = {STANDARD_R}\n" + f"T = {STANDARD_T}\n"
    )
    for reac_id, cobrak_reaction in cobrak_model.reactions.items():
        if reac_id.endswith(cobrak_model.fwd_suffix):
            reverse_id = reac_id.replace(
                cobrak_model.fwd_suffix, cobrak_model.rev_suffix
            )
        elif reac_id.endswith(cobrak_model.rev_suffix):
            reverse_id = reac_id.replace(
                cobrak_model.rev_suffix, cobrak_model.fwd_suffix
            )
        else:
            reverse_id = ""

        if reverse_id in nlp_results:
            reac_flux = nlp_results[reac_id] - nlp_results[reverse_id]
        else:
            reac_flux = nlp_results[reac_id]

        if reac_flux <= abs(1e-12):
            continue

        e_conc = e_concs.get(reac_id, 1.0)

        tellurium_string += _get_reaction_string_of_cobrak_reaction(
            cobrak_model=cobrak_model,
            reac_id=reac_id,
            cobrak_reaction=cobrak_reaction,
            e_conc=e_conc,
            met_concs=met_concs,
            reac_flux=reac_flux,
            nlp_results=nlp_results,
            kinetic_ignored_metabolites=cobrak_model.kinetic_ignored_metabolites,
            unoptimized_reactions=unoptimized_reactions,
        )

    cobrak_model = delete_orphaned_metabolites_and_enzymes(cobrak_model)

    for unsafe_met_id, metabolite in cobrak_model.metabolites.items():
        original_met_id = unsafe_met_id
        met_id = _get_numbersafe_id(unsafe_met_id)
        if met_id in met_concs:
            tellurium_string += (
                f"\nconst substanceOnly species {met_id} = {met_concs[met_id] * 1_000 / cell_density}"
                f"\n{met_id}_molar := {met_id} * {cell_density / 1_000}\n"
            )
        else:
            if ("x_" + original_met_id in nlp_results) and (
                metabolite.log_min_conc != metabolite.log_max_conc
            ):
                exp_conc = (
                    exp(nlp_results[LNCONC_VAR_PREFIX + original_met_id])
                    * 1_000
                    / cell_density
                )
                tellurium_string += (
                    f"\nsubstanceOnly species {met_id} = {exp_conc}"
                    f"\n{met_id}_molar := {met_id} * {cell_density / 1_000}"
                )
            else:
                prefix = "const " if original_met_id.endswith("_e") else ""
                tellurium_string += (
                    f"\n{prefix}substanceOnly species {met_id} = {exp(metabolite.log_min_conc) * 1_000 / cell_density}"
                    f"\n{met_id}_molar := {met_id} * {cell_density / 1_000}"
                )

    return tellurium_string



"""get_protein_mass_mapping.py

Functions for the generation of a model's mapping of its proteins and their masses.
"""

# IMPORTS SECTION #
import time
from typing import Any

import cobra
import requests

from .io import ensure_folder_existence, json_load, json_write, standardize_folder


# FUNCTIONS SECTION #
def uniprot_get_enzyme_molecular_weights(
    model: cobra.Model,
    cache_basepath: str,
    multiplication_factor: float = 1 / 1000,
) -> dict[str, float]:
    """Returns a JSON with a mapping of protein IDs as keys, and as values the protein mass in kDa.

    The protein masses are taken  from UniProt (retrieved using
    UniProt's REST API).

    Arguments
    ----------
    * model: cobra.Model ~ The model in the cobrapy format

    Output
    ----------
    A JSON file with the path project_folder+project_name+'_protein_id_mass_mapping.json'
    and the following structure:
    <pre>
    {
        "$PROTEIN_ID": $PROTEIN_MASS_IN_KDA,
        (...),
    }
    </pre>
    """

    # GET UNIPROT ID - PROTEIN MAPPING
    uniprot_id_protein_id_mapping: dict[str, list[str]] = {}
    for gene in model.genes:
        # Without a UniProt ID, no mass mapping can be found
        if "uniprot" not in gene.annotation:
            continue
        uniprot_id = gene.annotation["uniprot"]
        if uniprot_id in uniprot_id_protein_id_mapping:
            uniprot_id_protein_id_mapping[uniprot_id].append(gene.id)
        else:
            uniprot_id_protein_id_mapping[uniprot_id] = [gene.id]

    # GET UNIPROT ID<->PROTEIN MASS MAPPING
    uniprot_id_protein_mass_mapping: dict[str, float] = {}
    # The cache stored UniProt masses for already searched
    # UniProt IDs (each file in the cache folder has the name
    # of the corresponding UniProt ID). This prevents searching
    # UniProt for already found protein masses. :-)
    cache_basepath = standardize_folder(cache_basepath)
    ensure_folder_existence(cache_basepath)
    cache_filepath = f"{cache_basepath}_cache_uniprot_molecular_weights.json"
    try:
        cache_json = json_load(cache_filepath, Any)
    except Exception:
        cache_json = {}
    # Go through each batch of UniProt IDs (multiple UniProt IDs
    # are searched at once in order to save an amount of UniProt API calls)
    # and retrieve the amino acid sequences and using these sequences, their
    # masses.
    print("Starting UniProt ID<->Protein mass search using UniProt API...")
    uniprot_ids = list(uniprot_id_protein_id_mapping.keys())
    batch_size = 15
    batch_start = 0
    while batch_start < len(uniprot_ids):
        # Create the batch with all UniProt IDs
        prebatch = uniprot_ids[batch_start : batch_start + batch_size]
        batch = []
        # Remove all IDs which are present in the cache (i.e.,
        # which were searched for already).
        # The cache consists of pickled protein mass floats, each
        # onein a file with the name of the associated protein.
        for uniprot_id in prebatch:
            if uniprot_id not in cache_json:
                batch.append(uniprot_id)
            else:
                uniprot_id_protein_mass_mapping[uniprot_id] = cache_json[uniprot_id]
                # print(uniprot_id + ":", uniprot_id_protein_mass_mapping[uniprot_id])

        # If all IDs could be found in the cache, continue with the next batch.
        if len(batch) == 0:
            batch_start += batch_size
            continue

        # Create the UniProt query for the batch
        # With 'OR', all given IDs are searched, and subsequently in this script,
        # the right associated masses are being picked.
        query = " OR ".join(batch)
        uniprot_query_url = f"https://rest.uniprot.org/uniprotkb/search?query={query}&format=tsv&fields=accession,id,mass"
        print(f"UniProt batch search for: {query}")

        # Call UniProt's API :-)
        uniprot_data = requests.get(uniprot_query_url, timeout=1e6).text.split("\n")
        # Wait in order to cool down their server :-)
        time.sleep(2.0)

        # Read out the API-returned lines
        found_ids = []
        for line in uniprot_data[1:]:
            if not line:
                continue
            accession_id = line.split("\t")[0].lstrip().rstrip()
            entry_id = line.split("\t")[1].lstrip().rstrip()
            mass_string = line.split("\t")[2].lstrip().rstrip()
            try:
                # Note that the mass entry from UniProt uses a comma as a thousand separator, so it has to be removed before parsing
                mass = float(mass_string.replace(",", ""))
            except ValueError:  # We may also risk the entry is missing
                # print(f"No protein mass obtainable for protein ID {uniprot_id}")
                continue
            uniprot_id_protein_mass_mapping[accession_id] = float(mass)
            uniprot_id_protein_mass_mapping[entry_id] = float(mass)
            found_ids.extend((accession_id, entry_id))

        # Create the pickled cache files for the searched protein masses
        for uniprot_id in found_ids:
            if (
                uniprot_id in uniprot_id_protein_mass_mapping
            ):  # Takes into account that we may fail to obtain a UniProt ID
                cache_json[uniprot_id] = uniprot_id_protein_mass_mapping[uniprot_id]

        # Continue with the next batch :D
        batch_start += batch_size

    # Create the final protein ID <-> mass mapping
    protein_id_mass_mapping: dict[str, float] = {}
    for uniprot_id in list(uniprot_id_protein_mass_mapping.keys()):
        try:
            protein_ids = uniprot_id_protein_id_mapping[uniprot_id]
        except Exception:
            continue
        for protein_id in protein_ids:
            protein_id_mass_mapping[protein_id] = (
                uniprot_id_protein_mass_mapping[uniprot_id] * multiplication_factor
            )

    # Return protein mass list JSON :D
    json_write(cache_filepath, cache_json)
    return protein_id_mass_mapping



"""This module contains the most convenient ways to create new Model instances from COBRApy models."""

import tempfile

# IMPORT SECTION #
from copy import deepcopy

import cobra
from cobra.manipulation import remove_genes
from numpy import log

from .brenda_functionality import brenda_select_enzyme_kinetic_data_for_model
from .cobrapy_model_functionality import get_fullsplit_cobra_model
from .constants import (
    REAC_ENZ_SEPARATOR,
    REAC_FWD_SUFFIX,
    REAC_REV_SUFFIX,
    STANDARD_R,
    STANDARD_T,
)
from .dataclasses import (
    Enzyme,
    EnzymeReactionData,
    ExtraLinearConstraint,
    Metabolite,
    Model,
    Reaction,
)
from .equilibrator_functionality import (
    equilibrator_get_model_dG0_and_uncertainty_values,
)
from .io import (
    get_files,
    json_load,
    json_write,
    load_unannotated_sbml_as_cobrapy_model,
    standardize_folder,
)
from .sabio_rk_functionality import sabio_select_enzyme_kinetic_data_for_model
from .uniprot_functionality import uniprot_get_enzyme_molecular_weights
from .utilities import (
    combine_enzyme_reaction_datasets,
    delete_orphaned_metabolites_and_enzymes,
    get_full_enzyme_mw,
    parse_external_resources,
)


# DATACLASS-USING FUNCTIONS #
def delete_enzymatically_suboptimal_reactions_in_fullsplit_cobrapy_model(
    cobra_model: cobra.Model,
    enzyme_reaction_data: dict[str, EnzymeReactionData | None],
    enzyme_molecular_weights: dict[str, float],
    fwd_suffix: str = REAC_FWD_SUFFIX,
    rev_suffix: str = REAC_REV_SUFFIX,
    reac_enz_separator: str = REAC_ENZ_SEPARATOR,
    special_enzyme_stoichiometries: dict[str, dict[str, float]] = {},
) -> cobra.Model:
    """Removes enzymatically suboptimal reactions from a fullsplit COBRApy model.

    This function identifies and deletes reactions in a COBRApy model that are enzymatically suboptimal based on
    enzyme reaction data and molecular weights. I.e., it retains only the reactions with the minimum molecular weight
    to k_cat (MW/k_cat) ratio for each base reaction. "base" reaction stands for any originally identical reaction, e.g.,
    if there are somehow now multiple phosphoglucokinase (PGK) reactions due to an enzyme fullsplit, only one of these
    PGK variants will be retained in the returned model.

    Args:
        cobra_model (cobra.Model): The COBRA model from which suboptimal reactions will be removed.
        enzyme_reaction_data (dict[str, EnzymeReactionData | None]): A dictionary mapping reaction IDs to
            ```EnzymeReactionData``` objects or ```None``` if the data is missing.
        enzyme_molecular_weights (dict[str, float]): A dictionary mapping enzyme identifiers to their molecular weights.

    Returns:
        cobra.Model: The modified COBRA model with suboptimal reactions removed.
    """
    reac_ids: list[str] = [reaction.id for reaction in cobra_model.reactions]
    ignored_reac_ids: list[str] = []
    base_reacs_to_min_mw_by_k_cat: dict[str, tuple[str, float]] = {}
    for reac_id in reac_ids:
        if reac_enz_separator not in reac_id:
            continue
        if reac_id not in enzyme_reaction_data:
            ignored_reac_ids.append(reac_id)
            continue

        current_enzyme_reaction_data = enzyme_reaction_data[reac_id]
        if current_enzyme_reaction_data is None:
            ignored_reac_ids.append(reac_id)
            continue

        mw = 0.0
        for identifier in current_enzyme_reaction_data.identifiers:
            if reac_id in special_enzyme_stoichiometries:
                if identifier in special_enzyme_stoichiometries[reac_id]:
                    stoichiometry = special_enzyme_stoichiometries[reac_id][identifier]
                else:
                    stoichiometry = 1.0
            else:
                stoichiometry = 1.0
            mw += stoichiometry * enzyme_molecular_weights[identifier]
        k_cat = current_enzyme_reaction_data.k_cat
        mw_by_k_cat = mw / k_cat

        if reac_id.endswith(fwd_suffix):
            direction_addition = fwd_suffix
        elif reac_id.endswith(rev_suffix):
            direction_addition = rev_suffix
        else:
            direction_addition = ""
        base_id = reac_id.split(reac_enz_separator)[0] + direction_addition

        if (
            base_id not in base_reacs_to_min_mw_by_k_cat
            or mw_by_k_cat < base_reacs_to_min_mw_by_k_cat[base_id][1]
        ):
            base_reacs_to_min_mw_by_k_cat[base_id] = (reac_id, mw_by_k_cat)
    enz_reacs_to_keep = [entry[0] for entry in base_reacs_to_min_mw_by_k_cat.values()]
    reacs_to_delete = [
        reac_id
        for reac_id in reac_ids
        if (reac_enz_separator in reac_id)
        and (reac_id not in enz_reacs_to_keep)
        and (reac_id not in ignored_reac_ids)
    ]
    cobra_model.remove_reactions(reacs_to_delete)
    return cobra_model


def delete_enzymatically_suboptimal_reactions_in_cobrak_model(
    cobrak_model: Model,
) -> Model:
    """Delete enzymatically suboptimal reactions in a COBRA model, similar to the idea in sMOMENT/AutoPACMEN [1].

    This function processes each reaction in the provided COBRA model to
    determine if it is enzymatically suboptimal based on Molecular Weight by k_cat (MW/kcat).
    Suboptimal reactions are identified by comparing their MW/kcat value with that of other reactions
    sharing the same base identifier, retaining only those with the lowest MW/kcat.
    The function then removes these suboptimal reactions from the model and cleans up orphaned metabolites.

    - The function assumes that the 'enzyme_reaction_data' attribute of each reaction includes
      identifiers and k_cat information for enzyme-catalyzed reactions. If not, those reactions are skipped.
    - Reactions with identical base IDs (but different directional suffixes) are considered as variants of the same reaction.
    - After removing suboptimal reactions, the function calls `delete_orphaned_metabolites_and_enzymes` to clean up any orphaned metabolites and enzymes that may have been left behind.

    [1] https://doi.org/10.1186/s12859-019-3329-9

    Parameters:
        cobrak_model (cobra.Model): A COBRA model containing biochemical reactions.

    Returns:
        cobra.Model: The updated COBRA model after removing enzymatically suboptimal reactions.
    """
    reac_id_to_mw_by_kcat: dict[str, float] = {}
    reac_id_to_base_id: dict[str, str] = {}
    base_id_to_min_mw_by_kcat: dict[str, float] = {}
    for reac_id, reac_data in cobrak_model.reactions.items():
        if reac_data.enzyme_reaction_data is None:
            continue
        if reac_data.enzyme_reaction_data.identifiers == []:
            continue
        if reac_data.enzyme_reaction_data.identifiers == [""]:
            continue

        mw_by_kcat = (
            get_full_enzyme_mw(cobrak_model, reac_data)
            / reac_data.enzyme_reaction_data.k_cat
        )

        reac_id_to_mw_by_kcat[reac_id] = mw_by_kcat

        if reac_id.endswith(cobrak_model.fwd_suffix):
            direction_addition = cobrak_model.fwd_suffix
        elif reac_id.endswith(cobrak_model.rev_suffix):
            direction_addition = cobrak_model.rev_suffix
        else:
            direction_addition = ""
        base_id = reac_id.split(cobrak_model.reac_enz_separator)[0] + direction_addition

        reac_id_to_base_id[reac_id] = base_id
        if base_id not in base_id_to_min_mw_by_kcat:
            base_id_to_min_mw_by_kcat[base_id] = mw_by_kcat
        else:
            base_id_to_min_mw_by_kcat[base_id] = min(
                base_id_to_min_mw_by_kcat[base_id], mw_by_kcat
            )

    reacs_to_delete = [
        reac_id
        for reac_id, base_id in reac_id_to_base_id.items()
        if reac_id_to_mw_by_kcat[reac_id] != base_id_to_min_mw_by_kcat[base_id]
    ]
    for reac_to_delete in reacs_to_delete:
        del cobrak_model.reactions[reac_to_delete]

    return delete_orphaned_metabolites_and_enzymes(cobrak_model)


def get_cobrak_model_from_sbml_and_thermokinetic_data(
    sbml_path: str,
    extra_linear_constraints: list[ExtraLinearConstraint],
    dG0s: dict[str, float],
    dG0_uncertainties: dict[str, float],
    conc_ranges: dict[str, tuple[float, float]],
    enzyme_molecular_weights: dict[str, float],
    enzyme_reaction_data: dict[str, EnzymeReactionData | None],
    max_prot_pool: float,
    kinetic_ignored_metabolites: list[str],
    enzyme_conc_ranges: dict[str, tuple[float, float] | None] = {},
    do_model_fullsplit: bool = False,
    do_delete_enzymatically_suboptimal_reactions: bool = True,
    R: float = STANDARD_R,
    T: float = STANDARD_T,
    fwd_suffix: str = REAC_FWD_SUFFIX,
    rev_suffix: str = REAC_REV_SUFFIX,
    reac_enz_separator: str = REAC_ENZ_SEPARATOR,
    omitted_metabolites: list[str] = [],
    keep_parameter_refs: bool = False,
) -> Model:
    """Creates a COBRAk model from an SBML and given further thermokinetic (thermodynamic and enzymatic) data.

    This function constructs a `Model` by integrating thermokinetic data and additional constraints
    into an existing COBRA model. It allows for the specification of concentration ranges, enzyme molecular weights, and
    reaction data, among other parameters.

    Args:
        sbml_path (str): The SBML model to be converted.
        extra_linear_constraints (list[ExtraLinearConstraint]): Additional linear constraints to be applied to the model.
        dG0s (dict[str, float]): Standard Gibbs free energy changes for reactions.
        dG0_uncertainties (dict[str, float]): Uncertainties in the standard Gibbs free energy changes.
        conc_ranges (dict[str, tuple[float, float]]): Concentration ranges for metabolites.
        enzyme_molecular_weights (dict[str, float]): Molecular weights of enzymes.
        enzyme_reaction_data (dict[str, EnzymeReactionData | None]): Enzyme reaction data for reactions.
        max_prot_pool (float): Maximum protein pool constraint.
        kinetic_ignored_metabolites (list[str]): Metabolites to be ignored in kinetic calculations.
        enzyme_conc_ranges (dict[str, tuple[float, float] | None], optional): Concentration ranges for enzymes. Defaults to {}.
        do_model_fullsplit (bool, optional): Whether to perform a full split of the model. Defaults to True.
        do_delete_enzymatically_suboptimal_reactions (bool, optional): Whether to delete enzymatically suboptimal reactions. Defaults to True.
        R (float, optional): Universal gas constant. Defaults to STANDARD_R.
        T (float, optional): Temperature in Kelvin. Defaults to STANDARD_T.

    Raises:
        ValueError: If a concentration range for a metabolite is not provided and no default is set.

    Returns:
        Model: The constructed `Model` with integrated thermokinetic data and constraints.
    """
    cobra_model = cobra.io.read_sbml_model(sbml_path)

    if do_model_fullsplit:
        cobra_model = get_fullsplit_cobra_model(cobra_model)

    cobrak_model = Model(
        reactions={},
        metabolites={},
        enzymes={},
        max_prot_pool=max_prot_pool,
        extra_linear_constraints=extra_linear_constraints,
        kinetic_ignored_metabolites=kinetic_ignored_metabolites,
        R=R,
        T=T,
        fwd_suffix=fwd_suffix,
        rev_suffix=rev_suffix,
        reac_enz_separator=reac_enz_separator,
    )

    for metabolite in cobra_model.metabolites:
        if metabolite.id in omitted_metabolites:
            continue

        if metabolite.id in conc_ranges:
            min_conc = conc_ranges[metabolite.id][0]
            max_conc = conc_ranges[metabolite.id][1]
        elif "DEFAULT" in conc_ranges:
            min_conc = conc_ranges["DEFAULT"][0]
            max_conc = conc_ranges["DEFAULT"][1]
        else:
            print(f"ERROR: No concentration range for metabolite {metabolite.id}.")
            print("Fixes: 1) Set its specific range; 2) Set a 'DEFAULT' range.")
            raise ValueError

        cobrak_model.metabolites[metabolite.id] = Metabolite(
            log_min_conc=log(min_conc),
            log_max_conc=log(max_conc),
            annotation=metabolite.annotation,
            name=metabolite.name,
            formula="" if not metabolite.formula else metabolite.formula,
            charge=metabolite.charge,
        )

    for reaction in cobra_model.reactions:
        dG0 = dG0s.get(reaction.id)

        dG0_uncertainty = dG0_uncertainties.get(reaction.id)

        if reaction.id in enzyme_reaction_data:
            used_enzyme_reaction_data = enzyme_reaction_data[reaction.id]
            if (used_enzyme_reaction_data is not None) and (not keep_parameter_refs):
                used_enzyme_reaction_data.k_cat_references = []
                used_enzyme_reaction_data.k_m_references = {}
                used_enzyme_reaction_data.k_i_references = {}
                used_enzyme_reaction_data.k_a_references = {}
        else:
            used_enzyme_reaction_data = None

        cobrak_model.reactions[reaction.id] = Reaction(
            min_flux=reaction.lower_bound,
            max_flux=reaction.upper_bound,
            stoichiometries={
                metabolite.id: value
                for (metabolite, value) in reaction.metabolites.items()
                if metabolite.id not in omitted_metabolites
            },
            dG0=dG0,
            dG0_uncertainty=dG0_uncertainty,
            enzyme_reaction_data=used_enzyme_reaction_data,
            annotation=reaction.annotation,
            name=reaction.name,
        )

    cobra_gene_ids = [gene.id for gene in cobra_model.genes]
    for enzyme_id, molecular_weight in enzyme_molecular_weights.items():
        min_enzyme_conc = None
        max_enzyme_conc = None
        if enzyme_id in enzyme_conc_ranges:
            conc_range = enzyme_conc_ranges[enzyme_id]
            if conc_range is not None:
                min_enzyme_conc = conc_range[0]
                max_enzyme_conc = conc_range[1]
        if enzyme_id in cobra_gene_ids:
            name = cobra_model.genes.get_by_id(enzyme_id).id
            annotation = cobra_model.genes.get_by_id(enzyme_id).annotation
        else:
            name = ""
            annotation = {}
        cobrak_model.enzymes[enzyme_id] = Enzyme(
            molecular_weight=molecular_weight,
            min_conc=min_enzyme_conc,
            max_conc=max_enzyme_conc,
            name=name,
            annotation=annotation,
        )

    if do_delete_enzymatically_suboptimal_reactions:
        cobrak_model = delete_enzymatically_suboptimal_reactions_in_cobrak_model(
            cobrak_model
        )

    return cobrak_model


def get_cobrak_model_with_kinetic_data_from_sbml_model_alone(
    sbml_path: str,
    path_to_external_resources: str,
    folder_of_sabio_database: str,
    brenda_version: str,
    prefer_brenda: bool,
    base_species: str,
    max_prot_pool: float,
    conc_ranges: dict[str, tuple[float, float]],
    inner_to_outer_compartments: list[str],
    phs: dict[str, float],
    pmgs: dict[str, float],
    ionic_strenghts: dict[str, float],
    potential_differences: dict[tuple[str, str], float],
    kinetic_ignored_enzymes: list[str],
    custom_kms_and_kcats: dict[str, EnzymeReactionData | None],
    kinetic_ignored_metabolites: list[str],
    do_model_fullsplit: bool = True,
    do_delete_enzymatically_suboptimal_reactions: bool = True,
    ignore_dG0_uncertainty: bool = True,
    enzyme_conc_ranges: dict[str, tuple[float, float] | None] = {},
    dG0_exclusion_prefixes: list[str] = [],
    dG0_exclusion_inner_parts: list[str] = [],
    dG0_corrections: dict[str, float] = {},
    extra_linear_constraints: list[ExtraLinearConstraint] = [],
    data_cache_folder: str = "",
    R: float = STANDARD_R,
    T: float = STANDARD_T,
    keep_parameter_refs: bool = False,
    enzymes_to_delete: list[str] = [],
    max_taxonomy_level: float = float("inf"),
) -> Model:
    """This functions creates a Model out of an SBML model and automatically collects kinetic and thermodynamic data.

    To make this function work, the functions needs in the COBRApy model...
    ...for the reactions: An EC number annotation [for enzymatic constraints]
    ...for the metabolites: BiGG IDs
    ...and all the other non-optional data in its arguments

    Args:
        sbml_path (str): The SBML model file path from which the Model and associated data retrieval shall be made
        path_to_external_resources (str): _description_
        brenda_version (str): _description_
        base_species (str): _description_
        max_prot_pool (float): _description_
        conc_ranges (dict[str, tuple[float, float]]): _description_
        inner_to_outer_compartments (list[str]): _description_
        phs (dict[str, float]): _description_
        pmgs (dict[str, float]): _description_
        ionic_strenghts (dict[str, float]): _description_
        potential_differences (dict[tuple[str, str], float]): _description_
        kinetic_ignored_enzymes (list[str]): _description_
        custom_kms_and_kcats (dict[str, EnzymeReactionData  |  None]): _description_
        kinetic_ignored_metabolites (list[str]): _description_
        do_model_fullsplit (bool, optional): _description_. Defaults to True.
        do_delete_enzymatically_suboptimal_reactions (bool, optional): _description_. Defaults to True.
        ignore_dG0_uncertainty (bool, optional): _description_. Defaults to True.
        enzyme_conc_ranges (dict[str, tuple[float, float]  |  None], optional): _description_. Defaults to {}.
        dG0_exclusion_prefixes (list[str], optional): _description_. Defaults to [].
        dG0_exclusion_inner_parts (list[str], optional): _description_. Defaults to [].
        extra_linear_constraints (list[ExtraLinearConstraint], optional): _description_. Defaults to [].
        data_cache_folder (str, optional): _description_. Defaults to "".
        R (float, optional): _description_. Defaults to STANDARD_R.
        T (float, optional): _description_. Defaults to STANDARD_T.

    Returns:
        Model: _description_
    """
    cobra_model = load_unannotated_sbml_as_cobrapy_model(sbml_path)
    remove_genes(
        model=cobra_model,
        gene_list=enzymes_to_delete,
        remove_reactions=False,
    )

    path_to_external_resources = standardize_folder(path_to_external_resources)
    if data_cache_folder:
        data_cache_folder = standardize_folder(data_cache_folder)
        data_cache_files = get_files(data_cache_folder)

    parse_external_resources(path_to_external_resources, brenda_version)

    if do_model_fullsplit:
        fullsplit_model = get_fullsplit_cobra_model(cobra_model)
    else:
        fullsplit_model = deepcopy(cobra_model)

    enzyme_reaction_data: dict[str, EnzymeReactionData | None] = {}
    if (not data_cache_folder) or (
        (data_cache_folder)
        and (
            ("_cache_dG0.json" not in data_cache_files)
            or ("_cache_dG0_uncertainties.json" not in data_cache_files)
        )
    ):
        brenda_enzyme_reaction_data = brenda_select_enzyme_kinetic_data_for_model(
            cobra_model=fullsplit_model,
            brenda_json_targz_file_path=f"{path_to_external_resources}brenda_2024_1.json.tar.gz",
            bigg_metabolites_json_path=f"{path_to_external_resources}bigg_models_metabolites.json",
            brenda_version=brenda_version,
            base_species=base_species,
            ncbi_parsed_json_path=f"{path_to_external_resources}parsed_taxdmp.json",
            kinetic_ignored_metabolites=kinetic_ignored_metabolites,
            kinetic_ignored_enzyme_ids=kinetic_ignored_enzymes,
            custom_enzyme_kinetic_data=custom_kms_and_kcats,
            max_taxonomy_level=max_taxonomy_level,
        )
        sabio_enzyme_reaction_data = sabio_select_enzyme_kinetic_data_for_model(
            cobra_model=fullsplit_model,
            sabio_target_folder=folder_of_sabio_database,
            base_species=base_species,
            ncbi_parsed_json_path=f"{path_to_external_resources}parsed_taxdmp.json",
            bigg_metabolites_json_path=f"{path_to_external_resources}bigg_models_metabolites.json",
            kinetic_ignored_metabolites=kinetic_ignored_metabolites,
            kinetic_ignored_enzyme_ids=kinetic_ignored_enzymes,
            custom_enzyme_kinetic_data=custom_kms_and_kcats,
            max_taxonomy_level=max_taxonomy_level,
        )

        enzyme_reaction_data = combine_enzyme_reaction_datasets(
            [
                (
                    brenda_enzyme_reaction_data
                    if prefer_brenda
                    else sabio_enzyme_reaction_data
                ),
                (
                    sabio_enzyme_reaction_data
                    if prefer_brenda
                    else brenda_enzyme_reaction_data
                ),
            ]
        )

        if data_cache_folder:
            json_write(
                f"{data_cache_folder}_cache_enzyme_reaction_data.json",
                enzyme_reaction_data,
            )
    else:
        enzyme_reaction_data = json_load(
            f"{data_cache_folder}_cache_enzyme_reaction_data.json",
            dict[str, EnzymeReactionData | None],
        )

    enzyme_molecular_weights = uniprot_get_enzyme_molecular_weights(
        model=fullsplit_model,
        cache_basepath=data_cache_folder,
    )

    if do_delete_enzymatically_suboptimal_reactions:
        fullsplit_model = (
            delete_enzymatically_suboptimal_reactions_in_fullsplit_cobrapy_model(
                fullsplit_model,
                enzyme_reaction_data,
                enzyme_molecular_weights,
            )
        )

    if (not data_cache_folder) or (
        (data_cache_folder)
        and (
            ("_cache_dG0.json" not in data_cache_files)
            or ("_cache_dG0_uncertainties.json" not in data_cache_files)
        )
    ):
        with tempfile.TemporaryDirectory() as tmpdict:
            cobra.io.write_sbml_model(fullsplit_model, tmpdict + "temp.xml")
        dG0s, dG0_uncertainties = equilibrator_get_model_dG0_and_uncertainty_values(
            tmpdict + "temp.xml",
            inner_to_outer_compartments,
            phs,
            pmgs,
            ionic_strenghts,
            potential_differences,
            dG0_exclusion_prefixes,
            dG0_exclusion_inner_parts,
            ignore_dG0_uncertainty,
        )
        if data_cache_folder:
            json_write(f"{data_cache_folder}_cache_dG0.json", dG0s)
            json_write(
                f"{data_cache_folder}_cache_dG0_uncertainties.json", dG0_uncertainties
            )
    else:
        dG0s = json_load(f"{data_cache_folder}_cache_dG0.json", dict[str, float])
        dG0_uncertainties = json_load(
            f"{data_cache_folder}_cache_dG0_uncertainties.json",
            dict[str, float],
        )

        dG0_keys = list(dG0s.keys())
        for dG0_key in dG0_keys:
            if any(
                dG0_key.startswith(dG0_exclusion_prefix)
                for dG0_exclusion_prefix in dG0_exclusion_prefixes
            ) or any(
                dG0_exclusion_inner_part in dG0_key
                for dG0_exclusion_inner_part in dG0_exclusion_inner_parts
            ):
                del dG0s[dG0_key]
                if dG0_key in dG0_uncertainties:
                    del dG0_uncertainties[dG0_key]

    for key, value in dG0_corrections.items():
        dG0s[key] += value

    with tempfile.TemporaryDirectory() as tmpdict:
        cobra.io.write_sbml_model(fullsplit_model, tmpdict + "temp.xml")
        return get_cobrak_model_from_sbml_and_thermokinetic_data(
            sbml_path=tmpdict + "temp.xml",
            extra_linear_constraints=extra_linear_constraints,
            dG0s=dG0s,
            dG0_uncertainties=dG0_uncertainties,
            conc_ranges=conc_ranges,
            enzyme_molecular_weights=enzyme_molecular_weights,
            enzyme_reaction_data=enzyme_reaction_data,
            max_prot_pool=max_prot_pool,
            kinetic_ignored_metabolites=kinetic_ignored_metabolites,
            enzyme_conc_ranges=enzyme_conc_ranges,
            R=R,
            T=T,
            keep_parameter_refs=keep_parameter_refs,
        )



"""Contains all dataclasses (and enums) used by COBRAk to define a metabolic model and its extra constraints and optimization objective.

Dataclasses are similar to structs in C: They are not intended to have member functions, only other types of member variables.
The main dataclass used by COBRAk is Model, which contains the full information about the metabolic model. As member variables,
a Model contains further dataclasses (such as Reaction, Metabolite, ...).
As dataclass_json is also invoked, it is possible to store and load the COBRAk dataclasses as JSON.
"""

# IMPORT SECTION #
from copy import deepcopy
from math import log
from typing import Any, Literal, TypeAlias

from pydantic import Field, FiniteFloat, NonNegativeInt, PositiveFloat
from pydantic.dataclasses import dataclass

from .constants import (
    QUASI_INF,
    REAC_ENZ_SEPARATOR,
    REAC_FWD_SUFFIX,
    REAC_REV_SUFFIX,
    STANDARD_R,
    STANDARD_T,
)


# DATACLASSES SECTION #
@dataclass
class Enzyme:
    """Represents an enzyme in a metabolic model.

    Members:
        molecular_weight (float):
            The enzyme's molecular weight in kDa.
        min_conc (float | None):
            [Optional] If wanted, one can set a special minimal concentration
            for the enzyme.
            Defaults to None, i.e., no given concentration value (i.e., only the total
            enzyme pool is the limit).
        max_conc (float | None):
            [Optional] If wanted, one can set a special maximal concentration
            for the enzyme.
            Defaults to None, i.e., no given concentration value (i.e., only the total
            enzyme pool is the limit).
        annotation (dict[str, str | list[str]]):
            [Optional] Dictionary containing additional enzyme annotation,
            e.g., {"UNIPROT_ID": "b12345"}.
            Defaults to '{}'.
        name: str:
            [Optional] Colloquial name of enzyme
    """

    molecular_weight: FiniteFloat
    """The enzyme's molecular weight in kDa"""
    min_conc: PositiveFloat | None = None
    """[Optional] The enzyme's minimal concentration in mmol⋅gDW⁻¹"""
    max_conc: PositiveFloat | None = None
    """[Optional] The enzyme's minimal concentration in mmol⋅gDW⁻¹"""
    annotation: dict[str, str | list[str]] = Field(default_factory=dict)
    """[Optional] Any annotation data for the enzyme (e.g., references). Has no effect on calculations"""
    name: str = ""
    """Colloquial name of enzyme"""


@dataclass
class ParameterReference:
    """Represents the database reference for a kinetic parameter."""

    database: str = ""
    """(If given) The database from which this parameter was read. Defaults to ''."""
    comment: str = "(no refs)"
    """Any comment given for this value (e.g. literature)? Defaults to '(no refs)'."""
    species: str = ""
    """Scientific name of the species where this value was measured. Defaults to ''."""
    substrate: str = ""
    """The metabolite (or reaction substrate) for which this value was measured. Defaults to ''."""
    pubs: list[str] = Field(default_factory=list)
    """"""
    tax_distance: int | None = None
    value: float | None = None


@dataclass
class EnzymeReactionData:
    """Represents the enzymes used by a reaction."""

    identifiers: list[str]
    """The identifiers (must be given in the associated Model enzymes instance) of the reaction's enzyme(s)"""
    k_cat: PositiveFloat
    """The reaction's k_cat (turnover numbers) in h⁻¹"""
    k_cat_references: list[ParameterReference] = Field(default_factory=list)
    """[Optional] List of references showing the source(s) of the k_cat value"""
    k_ms: dict[str, PositiveFloat] = Field(default_factory=dict)
    """[Optional] The reaction's k_ms (Michaelis-Menten constants) in M=mol⋅l⁻¹. Metabolite IDs are keys, k_ms values. Default is {}"""
    k_m_references: dict[str, list[ParameterReference]] = Field(default_factory=dict)
    """[Optional] References showing the source(s) of the k_m values. Metabolite IDs are keys, the source lists values. Default is {}"""
    k_is: dict[str, PositiveFloat] = Field(default_factory=dict)
    """[Optional] The reaction's k_is (Inhibition constants) in M=mol⋅l⁻¹. Metabolite IDs are keys, k_ms values. Default is {}"""
    k_i_references: dict[str, list[ParameterReference]] = Field(default_factory=dict)
    """[Optional] References showing the source(s) of the k_i values. Metabolite IDs are keys, the source lists values. Default is {}"""
    k_as: dict[str, PositiveFloat] = Field(default_factory=dict)
    """[Optional] The reaction's k_as (Activation constants) in M=mol⋅l⁻¹. Metabolite IDs are keys, k_ms values. Default is {}"""
    k_a_references: dict[str, list[ParameterReference]] = Field(default_factory=dict)
    """[Optional] References showing the source(s) of the k_a values. Metabolite IDs are keys, the source lists values. Default is {}"""
    special_stoichiometries: dict[str, PositiveFloat] = Field(default_factory=dict)
    """[Optional] Special (non-1) stoichiometries of polypeptides in the reaction's enzyme. Default is {}"""


@dataclass
class ExtraLinearConstraint:
    """Represents a general linear Model constraint.

    This can affect not only reactions, but also all other
    variables set in a COBRAk model. E.g., if one wants (for whatever
    reason) the following constraint:
    0.5 <= [A] - 2 * r_R1 <= 2.1
    the corresponding ExtraLinearConstraint instance would be:
    ExtraLinearConstraint(
        stoichiometries = {
            "x_A": 1.0,
            "R1": -2,
        },
        lower_value = 0.5,
        upper_value = 2.1,
    )
    lower_value or upper_value can be None if no such limit is desired.
    """

    stoichiometries: dict[str, float]
    """Keys: Model variable names; Children: Multipliers of constraint"""
    lower_value: float | None = None
    """Minimal numeric constraint value. Either this and/or upper_value must be not None. Defaults to None."""
    upper_value: float | None = None
    """Maximal numeric constraint value. Either this and/or lower_value must be not None. Defaults to None."""


@dataclass
class Metabolite:
    """Represents a Model's metabolite."""

    log_min_conc: FiniteFloat = log(1e-6)
    """Maximal logarithmic concentration (only relevant for thermodynamic constraints); Default is log(1e-6 M)"""
    log_max_conc: FiniteFloat = log(0.02)
    """Maximal logarithmic concentration (only relevant for thermodynamic constraints); Default is log(0.02 M)"""
    annotation: dict[str, str | list[str]] = Field(default_factory=dict)
    """Optional annotation (e.g., CHEBI numbers, ...); Default is {}"""
    name: str = ""
    """Colloquial name of metabolite"""
    formula: str = ""
    """Chemical formula of metabolite"""
    charge: int = 0
    """Electron charge of metabolite"""


@dataclass
class Reaction:
    """Represents a Model's reaction.

    E.g., a reaction
    A -> B [0; 1000], ΔG'°=12.1 kJ⋅mol⁻¹, catalyzed by E1 with k_cat=1000 h⁻¹
    would be
    Reaction(
        stoichiometries: {
            "A": -1,
            "B": +1,
        },
        min_flux: 0,
        max_flux: 1000,
        dG0=12.1,
        dG0_uncertainty=None,
        enzyme_reaction_data=EnzymeReactionData(
            identifiers=["E1"],
            k_cat=1000,
            k_ms=None,
            k_is=None,
            k_as=None,
        ),
        annotation={}, # Can be also ignored
    )
    """

    stoichiometries: dict[str, float]
    """Metabolite stoichiometries"""
    min_flux: float
    """Minimal flux (for COBRAk, this must be ≥ 0)"""
    max_flux: float
    """Maximal flux (must be >= min_flux)"""
    dG0: FiniteFloat | None = None
    """If given, the Gibb's free energy of the reaction (only relevant for thermodynamic constraints); Default is None"""
    dG0_uncertainty: FiniteFloat | None = None
    """If given, the Gibb's free energy's uncertainty (only relevant for thermodynamic constraints); Default is None"""
    enzyme_reaction_data: EnzymeReactionData | None = None
    """If given, enzymatic data (only relevant for enzymatic constraints); Default is None"""
    annotation: dict[str, str | list[str]] = Field(default_factory=dict)
    """Optional annotation (e.g., KEGG identifiers, ...)"""
    name: str = ""
    """Colloquial name of reaction"""


@dataclass
class Model:
    """Represents a metabolic model in COBRAk.

    This includes its Reaction instances (which define the reaction stoichiometries),
    its Metabolite instances (which are referenced in the mentioned stoichiometries),
    as well as optional enzymatic and thermodynamic data.
    """

    metabolites: dict[str, Metabolite]
    """Keys: Metabolite IDs; Children: Metabolite instances"""
    reactions: dict[str, Reaction]
    """Keys: Reaction IDs; Children: Reaction instances"""
    enzymes: dict[str, Enzyme] = Field(default_factory=dict)
    """[Only neccessary with enzymatic constraints] Keys: Enzyme IDs; Children: Enzyme instances; default is {}"""
    max_prot_pool: PositiveFloat = Field(default=1e9)
    """[Only neccessary with enzymatic constraints] Maximal usable protein pool in g/gDW; default is 1e9, i.e. basically unrestricted"""
    extra_linear_constraints: list[ExtraLinearConstraint] = Field(default_factory=list)
    """[Optional] Extra linear constraints"""
    kinetic_ignored_metabolites: list[str] = Field(default_factory=list)
    """[Optional and only works with saturation term constraints] Metabolite IDs for which no k_m is neccessary"""
    R: PositiveFloat = Field(default=STANDARD_R)
    """[Optional and only works with thermodynamic constraints] Gas constant reference for dG'° in kJ⋅K⁻¹⋅mol⁻¹; default is STANDARD_R"""
    T: PositiveFloat = Field(default=STANDARD_T)
    """[Optional and only works with thermodynamic constraints] Temperature reference for dG'° in K; default is STANDARD_T"""
    annotation: dict[str, str | list[str]] = Field(default_factory=dict)
    """[Optional] Any annotation for the model itself (e.g., its name or references). Has no effect on calculations."""
    reac_enz_separator: str = REAC_ENZ_SEPARATOR
    """[Optional] String infix that separated reaction IDs of reaction with multiple enzyme variants from their enzyme ID. Defaults to '_ENZ_'"""
    fwd_suffix: str = REAC_FWD_SUFFIX
    """[Optional] Reaction ID suffix of forward reaction variants (e.g. in a reversible reaction A→B, for the direction A→B). Default is '_FWD'"""
    rev_suffix: str = REAC_REV_SUFFIX
    """[Optional] Reaction ID suffix of reverse reaction variants (e.g. in a reversible reaction A→B, for the direction B→A). Default is '_REV'"""
    max_conc_sum: float = float("inf")
    """[Optional and only works with thermodynamic constraints] Maximal allowed sum of concentrations (for MILPs: linear approximation; for NLPs: Exact value). Inactive if set to default value of float('inf')"""
    conc_sum_ignore_prefixes: list[str] = Field(default_factory=list)
    """[Optional and only works with thermodynamic constraints] """
    conc_sum_include_suffixes: list[str] = Field(default_factory=list)
    """[Optional and only works with thermodynamic constraints] """
    conc_sum_max_rel_error: float = 0.05
    """[Optional and only works with MILPs with thermodynamic constraints] Maximal relative concentration sum approximation error"""
    conc_sum_min_abs_error: float = 1e-6
    """[Optional and only works with MILPs with thermodynamic constraints] Maximal absolute concentration sum approximation error"""

    def __enter__(self):
        """Method called when entering 'with' blocks"""
        # Return a deep copy of self
        return deepcopy(self)

    def __exit__(self, a, b, c):
        """Method called when leaving a 'with' block"""
        # Cleanup can be performed here if needed
        return  # Return None to propagate any exceptions


@dataclass
class CorrectionConfig:
    """Stores the configuration for corrections in a model (see parameter corrections chapter in documentation)."""

    error_scenario: dict[str, tuple[float, float]] = Field(default_factory=list)
    """A dictionary where keys are error scenarios and values are tuples representing the lower and upper bounds of the error. Defaults to {}."""
    add_flux_error_term: bool = False
    """Indicates whether to add flux error terms. Defaults to False."""
    add_met_logconc_error_term: bool = False
    """Indicates whether to add metabolite log concentration error terms. Defaults to False."""
    add_enzyme_conc_error_term: bool = False
    """Indicates whether to add enzyme concentration error terms. Defaults to False."""
    add_kcat_times_e_error_term: bool = False
    """Indicates whether to add k_cat ⋅ [E] error terms. Defaults to False."""
    kcat_times_e_error_cutoff: PositiveFloat = 1.0
    """The cutoff value for the k_cat ⋅ [E] error term. Defaults to 1.0."""
    max_rel_kcat_times_e_correction: PositiveFloat = QUASI_INF
    """Maximal relative correction for the k_cat ⋅ [E] error error term. Defaults to QUASI_INF."""
    add_dG0_error_term: bool = False
    """Indicates whether to add ΔG'° error terms. Defaults to False."""
    dG0_error_cutoff: PositiveFloat = 1.0
    """The cutoff value for the ΔG'° error terms. Defaults to 1.0."""
    max_abs_dG0_correction: PositiveFloat = QUASI_INF
    """Maximal absolute correction for the dG0 error term. Defaults to QUASI_INF."""
    add_km_error_term: bool = False
    """Indicates whether to add a kappa error term. Defaults to False."""
    km_error_cutoff: PositiveFloat = 1.0
    """Cutoff value for the κ error term. Defaults to 1.0."""
    max_rel_km_correction: PositiveFloat = 0.999
    """Maximal relative correction for the κ error term. Defaults to 0.999."""
    error_sum_as_qp: bool = False
    """Indicates whether to use a quadratic programming approach for the error sum. Defaults to False."""
    add_error_sum_term: bool = True
    """Whether to add an error sum term. Defaults to True."""
    use_weights: bool = False
    """Indicates whether to use weights for the corrections (otherwise, the weight is 1.0). Defaults to False."""
    weight_percentile: NonNegativeInt = 90
    """Percentile to use for weight calculation. Defaults to 90."""
    extra_weights: dict[str, float] = Field(default_factory=dict)
    """Dictionary to store extra weights for specific corrections. Defaults to {}."""
    var_lb_ub_application: Literal["", "exp", "log"] = ""
    """The application method for variable lower and upper bounds. Either '' (x=x), 'exp' or 'log'. Defaults to ''."""


@dataclass
class Solver:
    """Represents options for a pyomo-compatible solver"""

    name: str
    """The solver's name. E.g. 'scip' for SCIP and 'cplex_direct' for CPLEX."""
    solver_options: dict[str, float | int | str] = Field(default_factory=dict)
    """[Optional] Options transmitted to the solver itself."""
    solver_attrs: dict[str, float | int | str] = Field(default_factory=dict)
    """[Optional] Options set on the solver object in pyomo."""
    solve_extra_options: dict[str, Any] = Field(default_factory=dict)
    """[Optional] Options set on pyomo's solve function."""


# SHORTHAND TYPE ALIASES
ErrorScenario: TypeAlias = dict[str, tuple[float, float]]
"""A COBRAk error scenario type alias for a ConfigurationConfig; Is dict[str, tuple[float, float]]"""
OptResult: TypeAlias = dict[str, float]
"""A COBRAk variability optimization result type alias; Is dict[str, float]"""
VarResult: TypeAlias = dict[str, tuple[float | None, float | None]]
"""A COBRAk variability result type alias; Is dict[str, tuple[float | None, float | None]]"""



import json
import os
from shutil import rmtree

import cobra
import pytest

from cobrak.uniprot_functionality import uniprot_get_enzyme_molecular_weights


def test_uniprot_get_enzyme_molecular_weights():
    # Create a test model
    model = cobra.Model("test_model")
    gene = cobra.Gene("gene1")
    gene.annotation = {"uniprot": "P12345"}
    model.genes.append(gene)
    reaction = cobra.Reaction("reaction1")
    reaction.gene_reaction_rule = "gene1"
    model.reactions.append(reaction)

    # Call the function
    cache_basepath = "test_cache"
    protein_id_mass_mapping = uniprot_get_enzyme_molecular_weights(
        model, cache_basepath
    )

    # Check that the function returns a dictionary
    assert isinstance(protein_id_mass_mapping, dict)

    # Check that the dictionary contains the expected protein ID
    assert "gene1" in protein_id_mass_mapping

    # Check that the protein mass is a float
    assert isinstance(protein_id_mass_mapping["gene1"], float)


def test_uniprot_get_enzyme_molecular_weights_cache():
    # Create a test model
    model = cobra.Model("test_model")
    gene = cobra.Gene("gene1")
    gene.annotation = {"uniprot": "P12345"}
    model.genes.append(gene)
    reaction = cobra.Reaction("reaction1")
    reaction.gene_reaction_rule = "gene1"
    model.reactions.append(reaction)

    # Create a cache file
    cache_basepath = "test_cache"
    cache_filepath = f"{cache_basepath}_cache_uniprot_molecular_weights.json"
    cache_json = {"P12345": 100.0}
    with open(cache_filepath, "w") as f:
        json.dump(cache_json, f)

    # Call the function
    protein_id_mass_mapping = uniprot_get_enzyme_molecular_weights(
        model, cache_basepath
    )

    # Check that the function returns a dictionary
    assert isinstance(protein_id_mass_mapping, dict)

    # Check that the dictionary contains the expected protein ID
    assert "gene1" in protein_id_mass_mapping

    # Check that the protein mass is a float
    assert isinstance(protein_id_mass_mapping["gene1"], float)


def test_uniprot_get_enzyme_molecular_weights_no_uniprot_id():
    # Create a test model
    model = cobra.Model("test_model")
    gene = cobra.Gene("gene1")
    model.genes.append(gene)
    reaction = cobra.Reaction("reaction1")
    reaction.gene_reaction_rule = "gene1"
    model.reactions.append(reaction)

    # Call the function
    cache_basepath = "test_cache"
    protein_id_mass_mapping = uniprot_get_enzyme_molecular_weights(
        model, cache_basepath
    )

    # Check that the function returns an empty dictionary
    assert protein_id_mass_mapping == {}


def test_uniprot_get_enzyme_molecular_weights_invalid_model():
    # Call the function with an invalid model
    cache_basepath = "test_cache"
    with pytest.raises(AttributeError):
        uniprot_get_enzyme_molecular_weights("invalid_model", cache_basepath)


def test_uniprot_get_enzyme_molecular_weights_invalid_cache_basepath():
    # Create a test model
    model = cobra.Model("test_model")
    gene = cobra.Gene("gene1")
    gene.annotation = {"uniprot": "P12345"}
    model.genes.append(gene)
    reaction = cobra.Reaction("reaction1")
    reaction.gene_reaction_rule = "gene1"
    model.reactions.append(reaction)


def test_uniprot_get_enzyme_molecular_weights_cleanup():
    # Create a test model
    model = cobra.Model("test_model")
    gene = cobra.Gene("gene1")
    gene.annotation = {"uniprot": "P12345"}
    model.genes.append(gene)
    reaction = cobra.Reaction("reaction1")
    reaction.gene_reaction_rule = "gene1"
    model.reactions.append(reaction)

    # Call the function
    cache_basepath = "test_cache"
    uniprot_get_enzyme_molecular_weights(model, cache_basepath)

    # Check that the cache file was created
    cache_filepath = f"{cache_basepath}_cache_uniprot_molecular_weights.json"
    assert os.path.exists(cache_filepath)

    # Clean up
    rmtree("./test_cache")
    os.remove("test_cache_cache_uniprot_molecular_weights.json")
